class BaseModel(object):
    def __init__(self, data):
        self._data = {}

        for key, value in data.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                self._data[key] = value

    @classmethod
    def parse(self, data):
        if isinstance(data, BaseModel):
            return data
        else:
            return self(data)


class ParentedObject(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent = None

    def set_parent(self, parent):
        self._parent = parent

    def get_parent(self, klass):
        current_obj = self
        while hasattr(current_obj, '_parent'):
            if isinstance(current_obj, klass):
                return current_obj
            else:
                current_obj = current_obj._parent

        if isinstance(current_obj, klass):
            return current_obj


class Extendable(object):
    def extend(self, other):
        raise NotImplementedError()



import sys
import pytest
import subprocess
from mock import MagicMock
from odoo_tools.odoo import Environment

from tests.utils import generate_addons


@pytest.fixture
def env(tmp_path):
    odoo_env = Environment()
    odoo_env.context.odoo_rc = tmp_path / 'odoo.cfg'
    yield odoo_env


@pytest.fixture
def addons_path(env, tmp_path):
    addons_path = tmp_path / 'addons'
    addons_path.mkdir(parents=True, exist_ok=True)

    generate_addons(addons_path, ['a', 'b', 'c'], depends=['base'])
    generate_addons(addons_path, ['d'], depends=['a'])

    env.context.custom_paths.add(addons_path)

    yield addons_path


def odoo_cleanup(env):
    print("tearing down")
    odoo_path = env.path()

    subprocess.run(['pip', 'uninstall', '-y', 'odoo'])

    odoo_modules = [
        key
        for key in sys.modules.keys()
        if key.startswith('odoo.')
    ]

    for mod in odoo_modules:
        del sys.modules[mod]

    subprocess.run(['rm', '-rf', str(odoo_path)])


@pytest.fixture(scope="module")
def odoo_env():
    env = Environment()

    options = MagicMock()

    options.languages = "fr_CA"
    options.upgrade = False
    options.target = False
    options.cache = False

    env.manage.install_odoo('15.0', opts=options)

    try:
        yield env
    finally:
        odoo_cleanup(env)


@pytest.fixture(scope="module")
def odoo_release():
    env = Environment()
    options = MagicMock()

    options.languages = "fr_CA"
    options.upgrade = False
    options.target = False
    options.cache = False

    env.manage.install_odoo('15.0', release="20220624", opts=options)

    try:
        yield env
    finally:
        odoo_cleanup(env)



import os
import pytest
from mock import patch

from odoo_tools.entrypoints import entrypoint
from odoo_tools.docker.user_entrypoint import (
    setup_env_config
)


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Testing Odoo is disabled"
)
def test_check_odoo(odoo_env):
    odoo_env.check_odoo()


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Testing Odoo is disabled"
)
def test_get_odoo_addons(odoo_env):
    assert len(odoo_env.addons_paths()) == 1

    bus_module = odoo_env.modules.get('bus')

    assert bus_module.technical_name == 'bus'

    from odoo.tools import config

    assert config == odoo_env.odoo_config()


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Testing Odoo is disabled"
)
def test_env_options(odoo_env):
    vals = dict(
        ODOO_LOAD="web,base,fun",
        ODOO_WORKERS="3",
        ODOO_DATABASE="fun"
    )

    with patch.dict(os.environ, vals):
        config = odoo_env.env_options()

        assert config['workers'] == "3"
        assert config['server_wide_modules'] == "web,base,fun"
        assert config['db_name'] == 'fun'


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Testing Odoo is disabled"
)
def test_release(odoo_release):
    from odoo.tools import config

    assert config == odoo_release.manage.config

    call_stack = []

    @entrypoint("odoo_tools.manage.before_config")
    def before_config(manage):
        call_stack.append('before_config')

    @entrypoint("odoo_tools.manage.initialize_odoo")
    def initialize_odoo(manage):
        call_stack.append("initialize_odoo")

    @entrypoint("odoo_tools.manage.after_config")
    def after_config(manage):
        call_stack.append("after_config")

    odoo_release.manage.initialize_odoo()

    assert call_stack == ["before_config", "initialize_odoo", "after_config"]


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Testing Odoo is disabled"
)
def test_init_db(odoo_release):
    db = odoo_release.manage.db("test12")

    db.default_entrypoints()

    odoo_release.manage.initialize_odoo()

    db.init(
        modules=["sale", "stock"],
        country="CA",
        language="fr_CA",
        without_demo=False
    )

    db.install_modules(["website"])

    db.uninstall_modules(["website", "stock"])

    with db.env() as env:
        IrModule = env['ir.module.module']
        modules = IrModule.search([['name', 'in', ['website', 'stock']]])
        for mod in modules:
            assert mod.state == 'uninstalled'

    website = odoo_release.modules.get('sale')

    website.export_translations(db, ['fr_CA'])


@pytest.mark.skipif(
    'TEST_ODOO' not in os.environ,
    reason="Requires odoo to fetch env_options"
)
def test_set_env_config(odoo_env):
    new_environ = {}

    with patch.dict(os.environ, new_environ, clear=True):
        setup_env_config(odoo_env)

    with odoo_env.config():
        assert odoo_env.get_config('db_host') is None

    new_environ = {
        "ODOO_DATABASE": "db1",
        "ODOO_LOAD": "web,base,kankan",
        "ODOO_WITHOUT_DEMO": "web",
    }

    with patch.dict(os.environ, new_environ, clear=True):
        setup_env_config(odoo_env)

    with odoo_env.config():
        assert odoo_env.get_config('db_name') == 'db1'
        assert odoo_env.get_config('server_wide_modules') == 'web,base,kankan'
        assert odoo_env.get_config('without_demo') == 'web'



import pytest
from odoo_tools.compat import Path
from odoo_tools.api.objects import Manifest
from odoo_tools.odoo import Environment


def generate_odoo_dir(tmp_path):
    odoo_dir = tmp_path / "odoo"
    odoo_dir.mkdir()
    addons_dir = odoo_dir / "addons"
    addons_dir.mkdir()

    return odoo_dir, addons_dir


def generate_addons(addons_dir, modules, **kw):
    for module in modules:
        mod_dir = addons_dir / module
        mod_dir.mkdir()

        manifest = mod_dir / '__manifest__.py'

        with manifest.open('w') as man:
            module_dict = {
                "description": module,
                "depends": ["web"]
            }

            for key, value in kw.items():
                module_dict[key] = value

            man.write(repr(module_dict))


def generate_files(current_folder, files):
    current_folder.mkdir(exist_ok=True, parents=True)

    for key, value in files.items():
        if isinstance(value, dict):
            generate_files(current_folder / key, value)
        else:
            cur_file = current_folder / key
            with cur_file.open("w") as fout:
                fout.write(value)


def generate_addons_full(addons_dir, module, manifest, files):
    module_path = addons_dir / module

    module_path.mkdir(exist_ok=True, parents=True)

    man = Manifest(
        module_path,
        manifest
    )
    man._manifest = man.path / '__manifest__.py'

    man.save()

    generate_files(man.path, files)



import click
import logging

from ..odoo import Environment

from .click.module import module
from .click.path import addons_paths
from .click.config import config
from .click.entrypoint import entrypoint
from .click.shell import shell
from .click.manage import manage
from .click.services import service
from .click.platform import platform


@click.group()
@click.option(
    '-c',
    '--config',
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
)
@click.option(
    '--exclude-odoo',
    is_flag=True,
    help="Exclude odoo paths from results",
    default=False
)
@click.option('--log-level')
@click.pass_context
def command(ctx, config, log_level, exclude_odoo):
    ctx.ensure_object(dict)
    env = Environment()

    if config:
        env.context.odoo_rc = config

    if exclude_odoo:
        env.context.exclude_odoo = exclude_odoo

    ctx.obj['env'] = env

    if log_level:
        logging.basicConfig(level=log_level)


command.add_command(config)
command.add_command(addons_paths)
command.add_command(module)
command.add_command(entrypoint)
command.add_command(shell)
command.add_command(manage)
command.add_command(service)
command.add_command(platform)



import click
import logging
from .utils import MODULE_TYPE
from ...compat import Path

from ...configuration.misc import DictObject


_logger = logging.getLogger(__name__)


@click.group()
def manage():
    pass


@manage.command(
    help="Initialize a database with modules and some configurations."
)
@click.argument("database")
@click.option(
    '-m',
    '--modules',
    type=MODULE_TYPE,
    help="Modules to install",
    multiple=True
)
@click.option(
    '--with-demo',
    help="Generate new database with demo",
    is_flag=True,
    default=False
)
@click.option(
    '--country',
    help="Country two letter code",
)
@click.option(
    '-l',
    '--language',
    help="Languages to use",
    multiple=True
)
@click.pass_context
def init(ctx, database, modules, country, language, with_demo):
    env = ctx.obj['env']
    env.check_odoo()
    manage = env.manage.db(database)

    manage.default_entrypoints()

    if not language:
        language = ['en_US']

    to_install = {
        mod
        for mods in modules
        for mod in mods
    }

    manage.init(
        to_install,
        country=country,
        without_demo=not with_demo,
        language=",".join(language)
    )

    return True


@manage.command(
    help="Update specified modules in a database."
)
@click.argument("database")
@click.option(
    '-m',
    '--modules',
    type=MODULE_TYPE,
    help="Modules to install",
    multiple=True
)
@click.pass_context
def update(ctx, database, modules):
    env = ctx.obj['env']
    env.check_odoo()
    manage = env.manage.db(database)

    manage.default_entrypoints()

    to_update = {
        mod
        for mods in modules
        for mod in mods
    }

    manage.install_modules(
        to_update,
        phase="update modules",
        event="update_modules"
    )

    return True


@manage.command(
    help="Install modules in the provided database"
)
@click.argument("database")
@click.option(
    '-m',
    '--modules',
    type=MODULE_TYPE,
    help="Modules to install",
    multiple=True
)
@click.option(
    '-f',
    '--force',
    help="Force module init",
    is_flag=True,
    default=False
)
@click.pass_context
def install(ctx, database, modules, force):
    env = ctx.obj['env']
    env.check_odoo()
    manage = env.manage.db(database)

    manage.default_entrypoints()

    to_install = {
        mod
        for mods in modules
        for mod in mods
    }

    manage.install_modules(
        to_install,
        phase="Installing Modules",
        force=force,
        event="install_modules"
    )

    return True


@manage.command(
    help="Uninstall specified modules from database"
)
@click.argument("database")
@click.option(
    '-m',
    '--modules',
    type=MODULE_TYPE,
    help="Modules to install",
    multiple=True
)
@click.pass_context
def uninstall(ctx, database, modules):
    env = ctx.obj['env']
    env.check_odoo()
    manage = env.manage.db(database)

    manage.default_entrypoints()

    to_remove = {
        mod
        for mods in modules
        for mod in mods
    }

    manage.uninstall_modules(
        to_remove
    )

    return True


@manage.command(
    help="Install Odoo in the current environment."
)
@click.argument(
    "version"
)
@click.option(
    '--release',
    help="Release version from https://nightly.odoo.com",
)
@click.option(
    '--repo',
    help="Repository",
    default="https://github.com/odoo/odoo.git"
)
@click.option(
    '--ref',
    help="Git reference to fetch the code",
)
@click.option(
    '--languages',
    help="Languages to keep",
    default="all"
)
@click.option(
    '--cache',
    help="Cache directory",
)
@click.option(
    '--target',
    help="Target path in which to install",
)
@click.option(
    '--upgrade',
    help="Upgrade pip packages while installing",
    is_flag=True,
    default=False
)
@click.pass_context
def setup(
    ctx,
    version,
    release,
    repo,
    ref,
    cache,
    languages,
    target,
    upgrade
):
    env = ctx.obj['env']

    opts = DictObject()

    if not languages:
        opts.languages = "all"
    else:
        opts.languages = languages

    if upgrade:
        opts.upgrade = True

    if target:
        opts.target = Path.cwd() / target

    env.manage.install_odoo(
        version,
        release=release,
        ref=ref or version,
        repo=repo,
        opts=opts
    )



"""
Render Modules
==============

This module cover a series of function that can be used to render
the html description of the module based on static content and the
description string in the manifest.

This behaves in similar ways as how Odoo render the description
string for modules within Odoo.

You may want to render html description of the module when you need
to display the module description to Odoo users or potentially in a
web store.
"""
import lxml
import lxml.html

from docutils import nodes
from docutils.utils import Reporter
from docutils.core import publish_string
from docutils.transforms import Transform, writer_aux
from docutils.writers.html4css1 import Writer


class MyFilterMessages(Transform):
    """
    Custom docutils transform to remove `system message` for a document and
    generate warnings.

    (The standard filter removes them based on some `report_level` passed in
    the `settings_override` dictionary, but if we use it, we can't see them
    and generate warnings.)
    """
    default_priority = 870

    def apply(self):
        for node in self.document.traverse(nodes.system_message):
            node.parent.remove(node)


class MyWriter(Writer):
    """
    Custom docutils html4ccs1 writer that doesn't add the warnings to the
    output document.
    """
    def get_transforms(self):
        return [MyFilterMessages, writer_aux.Admonitions]


def render_description_str(path, description=''):
    html_file = path / "static/description/index.html"

    if html_file.exists():
        doc = html_file.open('rb').read()
        html = lxml.html.document_fromstring(doc)
        for element, attribute, link, pos in html.iterlinks():
            # TODO convert to an actual url that can work
            # especially if the assets aren't loaded anywhere exactly
            if (
                element.get('src') and
                '//' not in element.get('src') and
                'static/' not in element.get('src')
            ):
                element.set(
                    'src',
                    "/%s/static/description/%s" % (
                        path.name, element.get('src')
                    )
                )

        return lxml.html.tostring(html).decode()

    readme_files = ['README.rst', 'README.md', 'README.txt']

    for readme in readme_files:
        readme_path = path / readme

        if not readme_path.exists():
            continue

        readme_data = readme_path.open().read()
        break
    else:
        readme_data = description

    overrides = {
        'embed_stylesheet': False,
        'doctitle_xform': False,
        'output_encoding': 'unicode',
        'xml_declaration': False,
        'file_insertion_enabled': False,
        'report_level': Reporter.ERROR_LEVEL
    }

    description_html = publish_string(
        source=readme_data,
        settings_overrides=overrides,
        writer=MyWriter()
    )

    return description_html



"""
Overview
========

This library came to light while developing entrypoints
for self configuring docker image. Some of the initial
functions were to search for odoo addons paths and search
for files like requirements.txt and apt-package.txt to
install python libraries the same way OdooSH_ does.

Then internally, some continuous integration tools shared
a lot of those functions that were initially duplicated in
both projects. Eventually, it became obvious that it would
be easier to have a single library that could be used by
both projects.

Then more things were added like the capacity to modify
manifest files programmatically, a shell using ptpython
to quickly enter a python shell with a properly
setup odoo environment, entrypoints from the docker image,
db api to initialize or manage an existing database,
translation tools to generate po and pot files.

This library can be used as is with the available command
lines to manage things manually, but its real purpose is
to be used in an automated environment.

And one thing that's particular is that this library
doesn't require odoo to function for most things. You
can use it to actually install odoo since odoo doesn't
provide easily installable pypi packages unfortunately.

.. _OdooSH: https://odoo.sh
"""



from odoo_tools.api.objects import get_translation_filename


def test_trans_filename():
    assert get_translation_filename('fr_CA', 'bus') == 'fr_CA.po'
    assert get_translation_filename('fr_FR', 'bus') == 'fr.po'
    assert get_translation_filename('', 'bus') == 'bus.pot'
    assert get_translation_filename(None, 'bus') == 'bus.pot'
    assert get_translation_filename(False, 'bus') == 'bus.pot'



"""
Environment Variables
=====================
"""
from os import environ
from .utils import from_csv


class EnvironmentVariables(object):
    """
    EnvironmentVariables parser

    Attributes:
        ODOO_BASE_PATH: The base path where odoo is installed

        ODOO_RC: Location of the odoo.cfg file

        ODOO_STRICT_MODE: Run odoo in Strict Mode

        ODOO_EXCLUDED_PATHS (list): Excluded paths will not be looked into when
            searching modules

        ODOO_EXTRA_PATHS (list): Extra paths are paths to look for modules other
            than the default ones.

        ODOO_EXTRA_APT_PACKAGES: Extra apt packages to install other than the one
            introspected in modules.

        ODOO_DISABLED_MODULES: Odoo modules that shouldn't exist. Those would get
            removed from the addons paths.

        MASTER_PASSWORD: Master password that should be set instead of the default
            one being automatically generated randomly.

        SHOW_MASTER_PASSWORD: Log the master password in the logs.

        SKIP_PIP: Skip the installation of pip modules.

        SKIP_SUDO_ENTRYPOINT: Skip the sudo entrypoint. The entry point is used
            mainly to setup things that require higher access like installing apt
            packages.

        SKIP_POSTGRES_WAIT: Tells to skip waiting for postgress to be up and
            running

        I_KNOW_WHAT_IM_DOING: This environment variables allows it to set the
            PGPASSWORD in the environment variables. Otherwise, the entrypoint
            will fail to let odoo starts unless you specify this environment
            variable as TRUE.  The reason behind that is that it's generally not
            a good idea to set credentials in environment variables. Those values
            can be leaked easily in logs from the environment running the container
            or even in logs displaying environment variables anywhere. This
            should only be used in dev environment or test environment in which
            credentials are not that important.

        ODOO_VERSION: The odoo version being currently used.

        RESET_ACCESS_RIGHTS: Reset access rights ensure that files in the home
            directory of the odoo user is set to the right user. This shouldn't
            be necessary in most cases.

        DEPLOYMENT_AREA: Deployment Area is just an environment variable that can
            be set to determine if it's a production environment or test
            environment.

        APT_INSTALL_RECOMMENDS: Define if you want the sudo_entrypoint to call
            apt-get with --no-install-recommends. By default, the entrypoint will
            use --no-install-recommends, but if this environment variable is set
            to TRUE. It will not add this parameter to apt-get install.

        ODOO_REQUIREMENTS_FILE: The path in which requirements get stored during
            installation of pip packages required by odoo addons.

        USE_ODOO_LOGGER: Tell the app to initialize the odoo logger instead of
            using the default one.

        PACKAGE_MAP_FILE: File storing a map of {module_name: module_renamed} to
            rename modules that shouldn't be used as is.
    """
    def __init__(self):
        self.ODOO_BASE_PATH = environ.get('ODOO_BASE_PATH')
        self.ODOO_RC = environ.get('ODOO_RC')
        self.ODOO_STRICT_MODE = environ.get('ODOO_STRICT_MODE')
        self.ODOO_EXCLUDED_PATHS = from_csv(
            environ.get('ODOO_EXCLUDED_PATHS', ''),
            set
        )
        self.ODOO_EXTRA_PATHS = from_csv(
            environ.get('ODOO_EXTRA_PATHS'),
            set
        )
        self.ODOO_EXTRA_APT_PACKAGES = from_csv(
            environ.get('ODOO_EXTRA_APT_PACKAGES') or
            environ.get('EXTRA_APT_PACKAGES'),
            set
        )
        self.ODOO_DISABLED_MODULES = from_csv(
            environ.get('ODOO_DISABLED_MODULES'),
            set
        )

        self.MASTER_PASSWORD = environ.get('MASTER_PASSWORD')
        self.SHOW_MASTER_PASSWORD = environ.get('SHOW_MASTER_PASSWORD')

        self.SKIP_PIP = environ.get('SKIP_PIP')
        self.SKIP_SUDO_ENTRYPOINT = environ.get('SKIP_SUDO_ENTRYPOINT')
        self.SKIP_POSTGRES_WAIT = environ.get('SKIP_POSTGRES_WAIT')
        self.I_KNOW_WHAT_IM_DOING = environ.get('I_KNOW_WHAT_IM_DOING')

        self.ALLOW_DANGEROUS_SETTINGS = environ.get(
            'I_KNOW_WHAT_IM_DOING',
            environ.get(
                'ALLOW_DANGEROUS_SETTINGS',
                ''
            )
        ).lower() == 'true'

        self.DEPLOYMENT_AREA = environ.get('DEPLOYMENT_AREA')

        self.ODOO_VERSION = environ.get('ODOO_VERSION')

        self.RESET_ACCESS_RIGHTS = environ.get('RESET_ACCESS_RIGHTS', '')

        self.APT_INSTALL_RECOMMENDS = environ.get('APT_INSTALL_RECOMMENDS')

        self.ODOO_REQUIREMENTS_FILE = environ.get('ODOO_REQUIREMENTS_FILE')

        self.USE_ODOO_LOGGER = environ.get('USE_ODOO_LOGGER')

        self.PACKAGE_MAP_FILE = environ.get('PACKAGE_MAP_FILE')

        self.REQUIREMENTS_FILE_PATH = environ.get('REQUIREMENTS_FILE_PATH')



import time
import os
from passlib.context import CryptContext

import logging

from ..utils import random_string
from ..compat import pipe, Path


_logger = logging.getLogger(__name__)


def run_command(params):
    """
    Main process running odoo
    """
    _logger.info("Starting main command: %s", params)

    if not params or len(params) == 0:
        return 0

    return pipe(params)


def call_sudo_entrypoint():
    """
    Call the odoo-entrypoint method with sudo.

    This method exist to ensure that a few things are well
    prepared for a docker image. It call the odoo-entrypoint
    again with root access to setup things like secrets,
    apt packages and so on.

    This can only be called once as the sudo access are stripped
    after calling this method once.

    Ideally, you'd want to prepare the odoo environment at build
    time for the image. Doing that at boot time consume time for
    nothing especially if you can build complete images that have
    all dependencies with pip/apt installed before executing the
    image.

    Entrypoints are just there to simplify setting up an image
    with addons mounted volumes that are only known at runtime.
    """
    command = ["sudo", "-H", "-E"]
    args = ["odootools", "entrypoint", "sudo"]

    ret = pipe(command + args)

    return ret


def install_python_dependencies(env):
    """
    Install all the requirements.txt file found
    """
    # TODO
    # https://pypi.org/project/requirements-parser/
    # to parse all the requirements file to parse all the possible specs
    # then append the specs to the loaded requirements and dump
    # the requirements.txt file in /var/lib/odoo/requirements.txt and
    # then install this only file instead of calling multiple time pip
    # all_paths = ['/addons'] + get_extra_paths()

    requirement_files = env.requirement_files()
    requirement_files_lst = list(set(requirement_files))
    requirement_files_lst.sort()

    message = "Installing python requirements found in:\n%s"

    _logger.info(
        message,
        "    \n".join([
            str(f_path)
            for f_path in requirement_files_lst
        ])
    )

    package_map = env.package_map()

    if env.context.requirement_file_path:
        file_path = env.context.requirement_file_path
    else:
        file_path = Path(env.context.odoo_rc).parent / 'requirements.txt'

    with file_path.open('w') as fout:
        requirements = list(env.modules.requirements(
            package_map=package_map,
            extra_paths=requirement_files,
        ))

        # Convert the set to a sorted list to prevent causing
        # changes with files with the same set but in a different
        # order.
        requirements.sort()

        data = "\n".join(requirements)

        fout.write(data)

    for requirements in requirement_files:
        _logger.info("Installing python packages from %s", requirements)

    _logger.info("Requirements:\n%s", data)

    retcode = pipe([
        "pip",
        "install",
        "--user",
        "-r", str(file_path)
    ])

    if env.context.strict_mode and retcode != 0:
        raise Exception("Failed to install pip dependencies")

    _logger.info("Installing python requirements complete")


def install_master_password(
    env,
    master_password_secret_path="/run/secrets/master_password"
):
    # Secure an odoo instance with a default master password
    # if required we can update the master password but at least
    # odoo doesn't get exposed by default without master passwords
    _logger.info("Installing master password in ODOORC")

    master_password_secret = Path(
        master_password_secret_path
    )

    if master_password_secret.exists():
        with master_password_secret.open("r") as mp:
            master_password = mp.read().strip()
    elif env.context.master_password:
        master_password = env.context.master_password
    else:
        master_password = random_string(64)

        if env.context.show_master_password:
            _logger.info(
                (
                    "Use this randomly generated master password"
                    " to manage the database\n"
                    "    %s"
                ),
                master_password
            )

    # Check that we don't have plaintext and encrypt it
    # This allow us to quickly setup servers without having to hash
    # ourselves first for security reason, you should always hash
    # the password first and not expect the image to do it correctly
    # but older version of odoo do not support encryption so only encrypt
    # older version of odoo...
    ctx = CryptContext(
        ['pbkdf2_sha512', 'plaintext'],
        deprecated=['plaintext']
    )
    if (
        env.odoo_version() > 10 and
        ctx.identify(master_password) == 'plaintext'
    ):
        master_password = ctx.encrypt(master_password)

    with env.config():
        env.set_config('admin_passwd', master_password)

    _logger.info("Installing master password completed")


def setup_addons_paths(env):
    with env.config():
        env.set_config('addons_path', ",".join([
            str(path) for path in env.addons_paths()
        ]))


def setup_server_wide_modules(env):
    _logger.info("Searching for server wide modules")

    server_wide_modules = env.modules.server_wide_modules()

    modules = ",".join(server_wide_modules)
    _logger.info("Setting server wide modules to %s" % (modules))
    with env.config():
        env.set_config('server_wide_modules', modules)


def setup_env_config(env):
    with env.config():
        for key, value in env.env_options().items():
            env.set_config(key, value)


def get_pg_environ(odoo_env, params):
    _logger.info("Configuring environment variables for postgresql")

    def ensure_config(env_key, param_long, param_small, odoo_config):
        """
        Check if config is in odoo_rc or command line
        """
        idx = -1
        value = False

        # First check for environment variable
        if env_key in os.environ:
            value = os.environ[env_key]

        # Then check for parameters
        if param_long and '--%s' % param_long in params:
            idx = params.index('--%s' % param_long)
            if not value:
                value = params[idx + 1] if idx < len(params) else None

        if param_small and '-%s' % param_small in params:
            idx = params.index('-%s' % param_small)
            if not value:
                value = params[idx + 1] if idx < len(params) else None

        # Then check config file
        if not value:
            value = odoo_env.get_config(odoo_config)

        return value, idx

    variables = [
        ('PGUSER', 'db_user', 'r', 'db_user'),
        ('PGHOST', 'db_host', None, 'db_host'),
        ('PGPORT', 'db_port', None, 'db_port'),
        ('PGDATABASE', 'database', 'd', 'db_name')
    ]

    # Accpet db_password only with this if some infra cannot be setup
    # otherwise...
    # It's a bad idea to pass password in cleartext in command line or
    # environment variables so please use .pgpass instead...
    if odoo_env.context.allow_dangerous_settings:
        variables.append(
            ('PGPASSWORD', 'db_password', 'w', 'db_password')
        )

    environ = {}

    # Setup basic PG env variables to simplify managements
    # combined with secret pg pass we can use psql directly
    new_params = params[:]

    with odoo_env.config():
        for env_key, param_long, param_small, odoo_config in variables:
            value, param_idx = ensure_config(
                env_key,
                param_long,
                param_small,
                odoo_config
            )

            if value:
                environ[env_key] = value
                odoo_env.set_config(odoo_config, value)

                # Override param
                if param_idx >= 0:
                    new_params[param_idx+1] = value

    _logger.info("Configuring environment variables done")

    return environ, new_params


def wait_postgresql(retry_count=None, retry_wait_time=None):
    import psycopg2

    if retry_count is None:
        retry_count = 5

    if retry_wait_time is None:
        retry_wait_time = 1

    error = None
    reset_db = False

    # Default database set to postgres
    if not os.environ.get('PGDATABASE'):
        os.environ['PGDATABASE'] = 'postgres'
        reset_db = True

    for retry in range(retry_count):
        try:
            _logger.info("Trying to connect to postgresql")
            # connect using defined env variables and pgpass files
            conn = psycopg2.connect("")
            message = "  Connected to %(user)s@%(host)s:%(port)s"
            _logger.info(message % conn.get_dsn_parameters())
            break
        except psycopg2.OperationalError as exc:
            error = exc
            time.sleep(retry_wait_time)
    else:
        # we reached the maximum retries so we trigger failure mode
        if error:
            _logger.info("Database connection failure %s", error)

        if reset_db:
            del os.environ['PGDATABASE']

        raise SystemError("Database connection failure")

    if reset_db:
        del os.environ['PGDATABASE']



import os
import pytest
from mock import patch, call, MagicMock
from odoo_tools.docker.user_entrypoint import (
    get_pg_environ,
    setup_env_config,
    setup_server_wide_modules,
    wait_postgresql,
    setup_addons_paths,
    install_master_password,
    install_python_dependencies,
    call_sudo_entrypoint,
    run_command,
)

from odoo_tools.docker.sudo_entrypoint import (
    install_apt_packages,
    fix_access_rights,
    remove_sudo
)
# from contextlib import nested
from odoo_tools.odoo import Environment
from odoo_tools.api.context import Context
from tests.utils import generate_addons


def test_get_pg_environ(tmp_path):
    env = Environment()
    env.context.odoo_rc = tmp_path / 'odoo.cfg'

    params = [
        '-d', 'db1',
        '-r', 'user1',
        '--db_host', 'postgres-1',
        '--db_port', '9090'
    ]

    new_environ = {
        "VAL": "1"
    }

    with patch.dict(os.environ, new_environ, clear=True):
        environ, params = get_pg_environ(env, params)

    assert environ['PGUSER'] == 'user1'
    assert environ['PGPORT'] == '9090'
    assert environ['PGHOST'] == 'postgres-1'
    assert environ['PGDATABASE'] == 'db1'

    with env.config():
        assert env.get_config('db_user') == 'user1'
        assert env.get_config('db_name') == 'db1'
        assert env.get_config('db_host') == 'postgres-1'
        assert env.get_config('db_port') == '9090'

    assert params == [
        '-d', 'db1',
        '-r', 'user1',
        '--db_host', 'postgres-1',
        '--db_port', '9090'
    ]


def test_get_pg_environ_env_priority(tmp_path):
    env = Environment()
    env.context.odoo_rc = tmp_path / 'odoo.cfg'

    params = [
        '-d', 'db1',
        '-r', 'user1',
        '--db_host', 'postgres-1',
        '--db_port', '9090'
    ]

    new_environ = {
        "PGHOST": 'postgres-2',
        "PGUSER": "user2",
        "PGDATABASE": "db2",
        "PGPORT": "9292",
    }

    with patch.dict(os.environ, new_environ, clear=True):
        environ, params = get_pg_environ(env, params)

    assert environ['PGUSER'] == 'user2'
    assert environ['PGPORT'] == '9292'
    assert environ['PGHOST'] == 'postgres-2'
    assert environ['PGDATABASE'] == 'db2'

    with env.config():
        assert env.get_config('db_user') == 'user2'
        assert env.get_config('db_name') == 'db2'
        assert env.get_config('db_host') == 'postgres-2'
        assert env.get_config('db_port') == '9292'

    assert params == [
        '-d', 'db2',
        '-r', 'user2',
        '--db_host', 'postgres-2',
        '--db_port', '9292'
    ]


def test_get_pg_environ_no_password(env):
    params = [
    ]

    new_environ = {
        "PGPASSWORD": "9292",
    }

    with patch.dict(os.environ, new_environ, clear=True):
        environ, params = get_pg_environ(env, params)

    assert 'PGPASSWORD' not in environ

    with env.config():
        assert env.get_config('db_password') is None

    assert params == []


def test_get_pg_environ_allow_password(env):
    env.context.allow_dangerous_settings = True

    params = [
    ]

    new_environ = {
        "PGPASSWORD": "9292",
    }

    with patch.dict(os.environ, new_environ, clear=True):
        environ, params = get_pg_environ(env, params)

    assert 'PGPASSWORD' in environ
    assert environ['PGPASSWORD'] == '9292'

    with env.config():
        assert env.get_config('db_password') == '9292'

    assert params == []


def test_get_pg_environ_no_params(tmp_path):
    env = Environment()
    env.context.odoo_rc = tmp_path / 'odoo.cfg'

    params = []

    new_environ = {}

    with patch.dict(os.environ, new_environ, clear=True):
        environ, params = get_pg_environ(env, params)

    assert environ == {}
    assert params == []


def test_wait_postgres_no_connect(env):
    from psycopg2 import OperationalError

    new_environ = {
    }

    # Defaults
    with patch.dict(os.environ, new_environ, clear=True), \
         patch('psycopg2.connect') as connect, \
         patch('time.sleep') as sleep:

        connect.side_effect = OperationalError
        sleep.return_value = None

        with pytest.raises(SystemError):
            wait_postgresql()

        assert connect.call_count == 5
        assert sleep.call_count == 5

        sleep.assert_has_calls([call(1)])

    # Custom params
    with patch.dict(os.environ, new_environ, clear=True), \
         patch('psycopg2.connect') as connect, \
         patch('time.sleep') as sleep:

        connect.side_effect = OperationalError
        sleep.return_value = None

        with pytest.raises(SystemError):
            wait_postgresql(4, 10)

        assert connect.call_count == 4
        assert sleep.call_count == 4

        sleep.assert_has_calls([call(10)])


def test_wait_postgres_connect(env):
    new_environ = {
    }

    with patch.dict(os.environ, new_environ, clear=True), \
         patch('psycopg2.connect') as connect, \
         patch('time.sleep') as sleep:

        conn_mock = MagicMock()
        conn_mock.get_dsn_parameters.return_value = {
            'user': '1',
            'host': 'pos',
            'port': '5224'
        }
        connect.return_value = conn_mock
        sleep.return_value = None

        wait_postgresql()

        assert connect.call_count == 1
        assert sleep.call_count == 0
        assert new_environ == {}


def test_setup_addons_paths(env, addons_path):
    with env.config():
        assert not env.get_config('addons_path')

    setup_addons_paths(env)

    with env.config():
        assert env.get_config('addons_path') == str(addons_path)


def test_server_wide_modules(env, addons_path):
    with env.config():
        assert not env.get_config('server_wide_modules')

    setup_server_wide_modules(env)

    with env.config():
        assert env.get_config('server_wide_modules') == "base,web"


def test_server_wide_module_extra(env, addons_path):
    generate_addons(addons_path, ['server_wide'], server_wide=True)

    setup_server_wide_modules(env)

    with env.config():
        assert env.get_config('server_wide_modules') == "base,web,server_wide"


def test_master_password_file_unencrypted(env, tmp_path):
    env.context.odoo_version = "8"

    mp = tmp_path / 'master_password'

    mp_password = "mypassword"

    with mp.open('w') as fout:
        fout.write(mp_password)

    new_environ = {}

    with patch.dict(os.environ, new_environ, clear=True):
        install_master_password(env, mp)

    with env.config():
        assert env.get_config('admin_passwd') == "mypassword"


def test_master_password_file_encrypted(env, tmp_path):
    env.context.odoo_version = "11"

    mp = tmp_path / 'master_password'

    mp_password = "mypassword"

    with mp.open('w') as fout:
        fout.write(mp_password)

    new_environ = {}

    with patch.dict(os.environ, new_environ, clear=True):
        install_master_password(env, mp)

    with env.config():
        assert env.get_config('admin_passwd') is not None
        assert env.get_config('admin_passwd') != mp_password


def test_master_password_env(env, tmp_path):
    env.context.odoo_version = "8"

    mp_password = "mypassword"

    new_environ = {
        "ODOO_VERSION": "8",
        "MASTER_PASSWORD": mp_password
    }

    with patch.dict(os.environ, new_environ, clear=True):
        env.context = Context.from_env()
        install_master_password(env)

    with env.config():
        assert env.get_config('admin_passwd') == mp_password


def test_master_password_random(env, tmp_path):
    env.context.odoo_version = "8"
    env.context.show_master_password = True

    new_environ = {
        "ODOO_VERSION": "8",
        "SHOW_MASTER_PASSWORD": "TRUE",
    }

    with patch.dict(os.environ, new_environ, clear=True), \
         patch('odoo_tools.docker.user_entrypoint.random_string') as random:

        random.return_value = "mocked"

        env.context = Context.from_env()
        install_master_password(env)

    with env.config():
        assert env.get_config('admin_passwd') == "mocked"
        random.assert_has_calls([call(64)])


def test_master_password_encrypted(env, tmp_path):
    env.context.odoo_version = "12"
    env.context.show_master_password = True

    new_environ = {
        "ODOO_VERSION": "12",
    }

    with patch.dict(os.environ, new_environ, clear=True), \
         patch('odoo_tools.docker.user_entrypoint.random_string') as random, \
         patch('odoo_tools.docker.user_entrypoint.CryptContext') as crypto:

        random.return_value = "mocked"

        crypto_context = MagicMock()

        crypto_context.identify.return_value = "plaintext"
        crypto_context.encrypt.return_value = "mocked_encrypted"

        crypto.return_value = crypto_context

        env.context = Context.from_env()
        install_master_password(env)

    with env.config():
        assert env.get_config('admin_passwd') == "mocked_encrypted"


def test_install_python_dependencies(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'
    env.context.custom_paths.add(tmp_path)

    addons_path = tmp_path / 'addons'
    addons_path.mkdir()

    deps = {
        "python": ['odoo-tools']
    }

    generate_addons(
        addons_path,
        ['server_wide'],
        external_dependencies=deps
    )

    with patch('odoo_tools.docker.user_entrypoint.pipe') as pipe:
        pipe.return_value = 0
        install_python_dependencies(env)
        pipe.assert_called_once()


def test_install_python_dependencies_exceptions(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'

    with patch('odoo_tools.docker.user_entrypoint.pipe') as pipe:
        pipe.return_value = 1
        with pytest.raises(Exception):
            install_python_dependencies(env)


def test_install_python_dependencies_exceptions_ignored(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'
    env.context.strict_mode = False

    with patch('odoo_tools.docker.user_entrypoint.pipe') as pipe:
        pipe.return_value = 1
        install_python_dependencies(env)
        pipe.assert_called_once()


def test_call_sudo_entrypoint(env):
    with patch('odoo_tools.docker.user_entrypoint.pipe') as pipe:
        pipe.return_value = 0

        ret = call_sudo_entrypoint()
        sudo_params = ['sudo', '-H', '-E', 'odootools', 'entrypoint', 'sudo']
        pipe.assert_has_calls([call(sudo_params)])
        assert ret == 0


def test_call_run_command(env):
    with patch('odoo_tools.docker.user_entrypoint.pipe') as pipe:
        pipe.return_value = 0
        params = ['odoo']
        ret = run_command(params)
        pipe.assert_has_calls([call(params)])
        assert ret == 0


def test_no_apt_packages(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'
    env.context.custom_paths.add(tmp_path)

    addons_path = tmp_path / 'addons'
    addons_path.mkdir()

    generate_addons(
        addons_path,
        ['server_wide'],
    )

    with patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:
        install_apt_packages(env)
        pipe.assert_not_called()


def test_apt_packages(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'
    env.context.custom_paths.add(tmp_path)

    addons_path = tmp_path / 'addons'
    addons_path.mkdir()

    apt_packages = addons_path / 'apt-packages.txt'

    with apt_packages.open('w') as fout:
        fout.write("git")

    generate_addons(
        addons_path,
        ['server_wide'],
    )

    with patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:
        pipe.return_value = 0
        install_apt_packages(env)
        assert pipe.call_count == 2


def test_apt_packages_exceptions(env, tmp_path):
    env.context.requirement_file_path = tmp_path / 'requirements.txt'
    env.context.custom_paths.add(tmp_path)

    addons_path = tmp_path / 'addons'
    addons_path.mkdir()

    apt_packages = addons_path / 'apt-packages.txt'

    with apt_packages.open('w') as fout:
        fout.write("git")

    generate_addons(
        addons_path,
        ['server_wide'],
    )

    with patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:
        pipe.side_effect = [1]
        with pytest.raises(SystemError):
            install_apt_packages(env)

        pipe.side_effect = [0, 1]
        with pytest.raises(SystemError):
            install_apt_packages(env)


def test_env_reset_access_rights(env, tmp_path):
    new_environ = {
        "RESET_ACCESS_RIGHTS": "",
    }

    with patch.dict(os.environ, new_environ, clear=True), \
         patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:

        env.context = Context.from_env()

        fix_access_rights(env)

        pipe.assert_not_called()

    new_environ = {
        "RESET_ACCESS_RIGHTS": "TRUE",
    }

    with patch.dict(os.environ, new_environ, clear=True), \
         patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:

        env.context = Context.from_env()

        fix_access_rights(env)

        assert pipe.call_count == 2


def test_remove_sudo(env, tmp_path):
    with patch('odoo_tools.docker.sudo_entrypoint.pipe') as pipe:
        remove_sudo()
        pipe.assert_called_once()



import pytest
import mock
from mock import patch, call
from pathlib import Path

from odoo_tools.utils import (
    to_path_list,
    is_subdir_of,
    filter_excluded_paths,
    convert_env_value,
    random_string
)


def test_to_path_list():
    lst = ['/a', 'b', '.c']
    pl = to_path_list(lst)

    for x in pl:
        assert isinstance(x, Path)

    pl = to_path_list([])
    assert len(pl) == 0
    assert isinstance(pl, list)


def test_is_subdir_of():
    a = Path('/a/b')
    b = Path('/a/b/c')
    c = Path('/b/c/d')

    assert is_subdir_of(a, b) is True
    assert is_subdir_of(b, a) is False
    # Non common root
    assert is_subdir_of(b, c) is False
    assert is_subdir_of(c, b) is False


def test_filter_excluded_paths():
    paths = [
        "/a",
        "/b",
        "/a/b",
        "/a/b/c",
        "/b/d/e",
    ]

    res = filter_excluded_paths(paths, to_path_list(['/b/d', '/a']))

    assert res == ['/b']


def test_convert_env_value():
    assert convert_env_value('test', 'True') is True
    assert convert_env_value('test', 'False') is False
    assert convert_env_value('test', 'FALSE') == 'FALSE'
    assert convert_env_value('test', 'TRUE') == 'TRUE'
    assert convert_env_value('test', '1') == '1'


def test_random_string():
    rand1 = random_string(10)
    rand2 = random_string(10)
    rand3 = random_string(10)
    rand4 = random_string(64)
    rand5 = random_string(64)
    rand6 = random_string(64)

    assert rand1 != rand2
    assert rand2 != rand3
    assert rand4 != rand5
    assert rand6 != rand3

    assert len(rand1) == 10
    assert len(rand2) == 10
    assert len(rand3) == 10

    assert len(rand4) == 64
    assert len(rand5) == 64
    assert len(rand6) == 64

    with patch('odoo_tools.utils.random') as random, \
         patch('odoo_tools.utils.range') as ranges:
        ranges.return_value = range(10)
        random.choice.return_value = '1'
        random_string(10)
        assert random.choice.call_count == 10
        ranges.assert_has_calls([call(10)])



import sys
import logging
from odoo_tools.compat import Path
import os
from contextlib import contextmanager
from six import ensure_str
from subprocess import run as sp_run, PIPE
from tempfile import TemporaryDirectory

_logger = logging.getLogger(__name__)


def find_in_path(binary, paths=None):
    """
    Find file in path

    Returns an absolute path to paths
    if the file is directly relative to '.' and
    if the file is present in paths.

    Otherwise the original path is returned.
    """
    binary_name = Path(binary)

    if binary_name.is_absolute():
        return binary

    if binary_name.parent.name != '':
        return binary

    for path in paths:
        found_binary = path / binary_name

        if found_binary.exists():
            return ensure_str(str(found_binary))

    return binary


def filter_args(args):
    """
    Take the first argument of the first args
    and filter the command line to modify the path if needed.
    """
    args = list(args)
    cmd, r_args = args[0], args[1:]
    program, options = cmd[0], cmd[1:]
    program = find_in_path(
        program,
        paths=[Path(sys.executable).parent]
    )
    cmd = [program] + options
    args = [cmd] + r_args

    return args


def run(*args, **kwargs):
    args = filter_args(args)

    if 'check' not in kwargs:
        kwargs['check'] = True

    ret = sp_run(*args, **kwargs)

    if kwargs.get('stdout') == PIPE:
        return ret.stdout

    return ret


@contextmanager
def cd(directory):
    cwd = Path.cwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(str(cwd))


def get_module_path(module):
    module_obj = sys.modules[module]
    return Path(module_obj.__file__).parent


def get_resource(module, path):
    module_path = get_module_path(module)
    return module_path / path


class DictObject(dict):
    def __getattr__(self, key):
        return self.get(key, False)

    def __setattr__(self, key, value):
        self[key] = value


def setup_logger():
    root_logger = logging.getLogger()
    log_level = os.environ.get('PYTHON_LOG', 'ERROR')
    root_logger.setLevel(log_level)



from odoo_tools.modules.render import render_description_str
from six import ensure_text


readme_file_data = """
README_FILE
===========

Some Readme File
"""


def test_rendering(tmp_path):
    description = render_description_str(tmp_path, "Empty")

    assert "Empty" in description


def test_static_rendering(tmp_path):
    html_file = tmp_path / "static/description/index.html"

    html_file.parent.mkdir(parents=True, exist_ok=True)

    assert tmp_path.parent.exists() is True

    with html_file.open("w") as fin:
        fin.write(ensure_text("STATIC_HTML_FILE"))

    description = render_description_str(tmp_path, "Empty")

    assert "STATIC_HTML_FILE" in description
    assert "Empty" not in description


def test_render_readme(tmp_path):
    readme_file = tmp_path / "README.rst"

    tmp_path.parent.mkdir(parents=True, exist_ok=True)

    with readme_file.open("w") as fin:
        fin.write(ensure_text(readme_file_data))

    description = render_description_str(tmp_path, "Empty")

    assert "README_FILE" in description



import os
from mock import patch
from odoo_tools.api.context import Context
from odoo_tools.configuration.misc import cd


def test_empty_context():
    context = Context()

    assert context.custom_paths == set()
    assert context.excluded_paths == set()
    assert context.extra_apt_packages == set()


def test_empty_context_odoorc(tmp_path):
    odoorc = tmp_path / 'odoo.cfg'

    with cd(tmp_path):
        context = Context()
        assert context.odoo_rc == odoorc


def test_context_envvars(tmp_path):

    excluded_paths = [
        tmp_path/'excluded1',
        tmp_path/'excluded2'
    ]

    extra_paths = [
        tmp_path / 'extra1',
        tmp_path / 'extra2'
    ]

    vals = dict(
        ODOO_EXTRA_PATHS=",".join([str(path) for path in extra_paths]),
        ODOO_STRICT_MODE="TRUE",
        ODOO_REQUIREMENTS_FILE=str(tmp_path/'requirements.txt'),
        ODOO_EXCLUDED_PATHS=",".join([str(path) for path in excluded_paths]),
        ODOO_DISABLED_MODULES="disabled1,disabled2",
        ODOO_RC=str(tmp_path / 'ODOORC')
    )

    with patch.dict(os.environ, vals):
        context = Context.from_env()

        assert context.custom_paths == set(extra_paths)
        assert context.odoo_rc == tmp_path / 'ODOORC'
        assert context.excluded_paths == set(excluded_paths)
        assert context.disabled_modules == {'disabled1', 'disabled2'}



import click
import sys
from ptpython.repl import embed

from ...compat import Path


@click.command()
@click.option(
    '-c',
    '--config',
    help="Config file"
)
@click.argument(
    'params',
    nargs=-1,
)
@click.pass_context
def shell(ctx, config, params):
    env = ctx.obj['env']
    try:
        import odoo
        from odoo.tools import config as odoo_config
        from odoo.api import Environment
        from contextlib import contextmanager, closing
        from odoo.release import version_info
    except ImportError:
        print("Odoo doesn't seem to be installed. Check your path")
        sys.exit(0)

    @contextmanager
    def manage():
        if version_info[0] >= 15:
            yield
        else:
            with Environment.manage():
                yield

    if config:
        odoo_config.config_file = str(Path.cwd() / config)
    else:
        odoo_config.config_file = env.context.odoo_rc

    odoo_config.parse_config(list(params))

    context = {}

    class ProtectedDict(dict):
        def __init__(self, protected_values):
            self.protected_values = protected_values

            super(ProtectedDict, self).__init__(
                **protected_values
            )

        def __setitem__(self, key, value):
            if key in self.protected_values:
                return
            super(ProtectedDict, self).__setitem__(key, value)

    with manage():
        registry = odoo.registry(odoo_config['db_name'])

        with closing(registry.cursor()) as cr:
            env = Environment(cr, 1, context)

            global_vals = {}
            local_vals = ProtectedDict({
                "env": env,
                "registry": registry,
                "cr": cr,
                "context": context
            })

            ret = embed(global_vals, local_vals)

    sys.exit(ret)



from odoo_tools.compat import Path
from ..configuration.git import checkout_repo
from .misc import (
    run,
    cd,
    get_resource
)
from tempfile import TemporaryDirectory
from six import ensure_binary, ensure_str
import logging
import requests
from packaging.version import parse as version_parse

_logger = logging.getLogger(__name__)


requirement_file_template = "requirements/requirements-{version}.0.txt"


class OdooSource(object):
    def __init__(self, version, options=None, cache=None):
        self.version = version
        self.cache = cache
        self.options = options
        self.target = None

    def __enter__(self):
        self.target = TemporaryDirectory()
        self.target_path = Path(self.target.name)

        if not self.cache:
            self.cache = self.target_path / 'cache'
            self.cache.mkdir(parents=True, exist_ok=True)

        return self

    def __exit__(self, *args, **kwargs):
        pass

    @property
    def parsed_version(self):
        return version_parse(self.version)

    @property
    def requirement_file(self):
        resource_file = requirement_file_template.format(
            version=self.parsed_version.major
        )
        resource_path = get_resource('odoo_tools', resource_file)
        if not resource_path.exists():
            resource_path = self.odoo_dir / 'requirements.txt'

        return resource_path

    def setup_installed_release(self, version, installed_release):
        _logger.info("Patching odoo/release.py with installed release")

        if version.major > 9:
            release_py = 'odoo/release.py'
        else:
            release_py = 'openerp/release.py'

        release_file = self.odoo_dir / release_py

        with release_file.open('r') as release_fd:
            release_code = release_fd.read()

        with release_file.open('wb') as release_fd:
            release_fd.write(ensure_binary(release_code))
            release_fd.write(ensure_binary("\n"))
            release_fd.write(ensure_binary(
                "installed_release = {}".format(installed_release)
            ))

    def strip_languages(self, languages):
        if not languages or languages == 'all':
            return

        prefix = set()

        for lang in languages.split(','):
            if not lang:
                continue

            lang_parts = lang.split('_')
            if len(lang_parts) == 2:
                prefix.add(lang_parts[0])
            prefix.add(lang)

        for file in self.odoo_dir.rglob('*.po'):
            filename = file.name[:-3]
            if filename not in prefix:
                try:
                    file.unlink()
                except Exception:
                    pass

    def fetch(self):
        raise NotImplementedError()

    def checkout(self):
        raise NotImplementedError()

    def need_update(self):
        try:
            import odoo
            release = odoo.release.installed_release
            if release == self.installed_release:
                _logger.info("Odoo is already up to date")
                return False
        except Exception:
            pass

        return True

    def install(self):
        self.setup_installed_release(
            self.parsed_version,
            self.installed_release
        )

        if hasattr(self.options, 'languages'):
            languages = self.options.languages
        else:
            languages = 'all'

        self.strip_languages(languages)

        _logger.info("Installing odoo")

        args = ['python', '-m', 'pip', 'install']

        if hasattr(self.options, 'upgrade') and self.options.upgrade:
            args.append('-U')

        if hasattr(self.options, 'target') and self.options.target:
            args += [
                "--target", str(self.options.target),
                "--implementation", "cp",
                "--python", "3.8",
                # "--only-binary", ":all:"
                "--no-deps",
            ]

        args += ['.', '-r', str(self.requirement_file)]

        with cd(self.odoo_dir):
            _logger.info("Installing odoo with command: '%s'", " ".join(args))
            run(args)


class GitRelease(OdooSource):
    def __init__(self, version, repo, ref, options=None, cache=None):
        super().__init__(version, options=options, cache=cache)
        self.repo = repo
        self.ref = ref

    @property
    def installed_release(self):
        return {
            "source": self.repo,
            "release": ensure_str(self.commit_id).replace('\n', ''),
        }

    def fetch(self):
        self.cached_odoo_dir = self.cache / 'odoo'
        self.cached_odoo_dir.mkdir(exist_ok=True, parents=True)

        with cd(self.cached_odoo_dir):
            head_file = self.cached_odoo_dir / 'HEAD'
            if not head_file.exists():
                run(['git', 'init', '--bare'])
                _logger.info("Fetching Odoo %:%", self.repo, self.ref)
                run(['git', 'fetch', '--depth', '1', self.repo, self.ref])
            else:
                run(['git', 'fetch', self.repo, self.ref])

    def checkout(self):
        self.odoo_dir = self.target_path / 'odoo'
        self.odoo_dir.mkdir(exist_ok=True, parents=True)

        self.commit_id = checkout_repo(self.cached_odoo_dir, self.odoo_dir)

    def install(self):
        target_addons_path = (
            'odoo/addons'
            if self.parsed_version.major > 9
            else 'openerp/addons'
        )

        target_path = self.odoo_dir / target_addons_path
        # Copy addons in /addons and add into odoo/addons to be installed
        # with the package
        _logger.info("Moving addons into package")
        for addons in (self.odoo_dir / 'addons').iterdir():
            run(['mv', str(addons), str(target_path)])

        super().install()


class OfficialRelease(OdooSource):

    def __init__(self, version, release, options=None, cache=None):
        super().__init__(version, options=options, cache=cache)
        self.release = release
        self.base_domain = "https://nightly.odoo.com"

    def fetch(self):

        path_fmt = "{version}/nightly/src"

        if '/' in self.release:
            release_path, release = self.release.split('/', 1)
            path = path_fmt.format(
                version=release_path
            )
        else:
            path = path_fmt.format(
                version=self.version
            )

        filename = "odoo_{version}.{release}.tar.gz".format(
            version=self.version,
            release=self.release
        )

        url = "{base_domain}/{path}/{filename}".format(
            base_domain=self.base_domain,
            path=path,
            filename=filename
        )

        output_file = self.cache / filename

        self.output_file = output_file

        if output_file.exists():
            _logger.info("Skipping fetch %s as file is already cached", url)
            return

        data_size = 0

        with requests.get(url, stream=True) as req:
            req.raise_for_status()

            with output_file.open('wb') as fout:
                for chunk in req.iter_content(chunk_size=8192):
                    data_size += len(chunk)
                    fout.write(chunk)

        _logger.info("Wrote % bytes to disk", data_size)

    def checkout(self):
        _logger.info("Extracting archive")
        output_dir = self.target_path / "odoo"
        output_dir.mkdir()
        run(['tar', '-xzf', str(self.output_file), '-C', str(output_dir)])
        self.odoo_dir = next(output_dir.iterdir())

    @property
    def installed_release(self):
        return {
            "source": self.base_domain,
            "release": self.release,
            "version": self.version
        }

    def install(self):
        super().install()
#
#
#   def setup_odoo_from_git(repo, ref, version, options=None):
#
#       if hasattr(options, 'cache'):
#           cache = options.cache
#       else:
#           cache = None
#
#       installer = GitRelease(
#           version,
#           repo,
#           ref,
#           options=options,
#           cache=cache
#       )
#
#       with installer:
#           if installer.need_update():
#               installer.fetch()
#               installer.checkout()
#               installer.install()
#
#
#   def setup_odoo_release(version, release, options=None):
#       if hasattr(options, 'cache'):
#           cache = options.cache
#       else:
#           cache = None
#
#       installer = OfficialRelease(
#           version,
#           release,
#           options=options,
#           cache=cache
#       )
#
#       with installer:
#           if installer.need_update():
#               installer.fetch()
#               installer.checkout()
#               installer.install()



import pytest
import os
from mock import patch
from cryptography.fernet import Fernet

from odoo_tools.configuration.git import ssh_command, fetch_addons
from odoo_tools.services.objects import ServiceManifests


@pytest.fixture
def services():
    fernet_key = Fernet.generate_key()

    fernet = Fernet(fernet_key)

    pk = fernet.encrypt(b"abcdef")

    services = {
        "services": [
            {
                "name": "base",
                "odoo": {
                    "version": "12.0"
                },
                "addons": [
                    {
                        "url": "git@github.com:oca/web.git"
                    },
                    {
                        "url": "git@github.com:oca/pivate.git",
                        "private_key": pk.decode()
                    },
                    {
                        "url": "git@github.com:oca/auth.git",
                        "auth": True,
                    }
                ]
            }
        ]
    }

    manifests = ServiceManifests.parse(services)

    return fernet_key, manifests


def test_ssh_command(services):
    fernet_key, manifests = services

    base_service = manifests.services['base']

    web_addons = base_service.addons['github_com_oca_web']
    private_addons = base_service.addons['github_com_oca_pivate']

    new_dict = {}

    with patch.dict(os.environ, new_dict, clear=True):
        with ssh_command(web_addons, decrypt_key=None):
            command = os.environ.get('GIT_SSH_COMMAND')
            assert command is None

        with ssh_command(private_addons, decrypt_key=fernet_key):
            command = os.environ.get('GIT_SSH_COMMAND')
            assert command.startswith('ssh -i')

        assert os.environ == {}


def test_fetch_addons(env, tmp_path, services):
    fernet_key, manifests = services

    fetch_dir = tmp_path / 'fetch_dir'

    new_dict = {}

    base_service = manifests.services['base']
    addon = base_service.addons['github_com_oca_web']

    with patch.dict(os.environ, new_dict, clear=True),\
         patch('odoo_tools.configuration.git.run') as run:

        run.return_value = "some_commit"

        with ssh_command(addon):
            path, res = fetch_addons(addon, fetch_dir)
            assert path.exists()
            assert isinstance(res, dict)
            assert res['url'] == addon.url
            assert res['commit'] == 'some_commit'


def tests_with_creds(env, tmp_path, services):
    fernet_key, manifests = services

    output_dir = tmp_path / 'fetch_dir'

    new_dict = {}

    base_service = manifests.services['base']
    addon = base_service.addons['github_com_oca_auth']

    with patch.dict(os.environ, new_dict, clear=True),\
         patch('odoo_tools.configuration.git.run') as run:

        run.return_value = "some_commit"

        creds = {
            'github.com': {
                'username': 'user',
                'password': 'password'
            }
        }

        fetch_url = "https://user:password@github.com/oca/auth.git"
        path, res = fetch_addons(addon, output_dir, credentials=creds)
        assert res['url'] == fetch_url



import click
from .utils import path_complete
from ...compat import Path


@click.group("path")
def addons_paths():
    pass


@addons_paths.command("ls")
@click.argument(
    'addons_path',
    nargs=-1,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    shell_complete=path_complete,
    required=False,
    # default=[]
)
@click.option(
    '--sorted',
    help="Sort paths",
    is_flag=True,
    default=False
)
@click.pass_context
def list_addons_paths(ctx, addons_path, sorted):
    env = ctx.obj['env']

    if addons_path:
        for path in addons_path:
            env.context.custom_paths.add(path)

        env.context.force_addons_lookup = True

    paths = env.addons_paths()

    if sorted:
        paths = list(paths)
        paths.sort()

    for path in paths:
        print(path)


@addons_paths.command("add")
@click.argument(
    'addons_path',
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.pass_context
def add_addons_path(ctx, addons_path):
    env = ctx.obj['env']

    env.context.custom_paths.add(addons_path)
    env.context.force_addons_lookup = True

    paths = [
        str(path)
        for path in env.addons_paths()
    ]

    with env.config() as config:
        config.set(
            'options',
            'addons_paths',
            ",".join(paths)
        )


@addons_paths.command("rm")
@click.argument(
    'addons_path',
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.pass_context
def remove_addons_path(ctx, addons_path):
    env = ctx.obj['env']

    remove_path = Path(addons_path)

    paths = [
        str(path)
        for path in env.addons_paths()
        if path != remove_path
    ]

    with env.config() as config:
        config.set(
            'options',
            'addons_paths',
            ",".join(paths)
        )



# import sys
from contextlib import closing, contextmanager
import logging
from itertools import groupby

from .objects import CompanySpec
from ..entrypoints import execute_entrypoint, entrypoint
from ..exceptions import InstallModulesError

_logger = logging.getLogger(__name__)


resettable_keys = [
    "load_language",
]


def set_missing_keys(config, key, value):
    try:
        if key not in config:
            config[key] = False
    except KeyError:
        config[key] = False


def before_config(man):
    man.config.config_file = man.environment.context.odoo_rc

    for key in resettable_keys:
        set_missing_keys(man.config, key, False)


def after_config(man):
    # Import models first to prevent
    # cannot import name 'Model' from partially initialized module
    # 'odoo.models'
    # (most likely due to a circular import)
    # import odoo # noqa
    # from odoo import models # noqa 
    # from odoo.osv import expression # noqa
    from odoo.service.server import load_server_wide_modules

    if hasattr(man, 'without_demo') and man.without_demo:
        man.config['without_demo'] = True
        for key in man.config['demo'].keys():
            man.config['demo'][key] = False

    if hasattr(man, 'languages') and man.languages:
        man.config['load_language'] = man.languages

    load_server_wide_modules()


def initialize_odoo(man):
    import odoo
    man.config._parse_config([])
    if man.environment.context.init_logger:
        odoo.netsvc.init_logger()
    man.config._warn_deprecated_options()
    odoo.modules.module.initialize_sys_path()


def ensure_db(man):
    import odoo
    from odoo.service.db import _create_empty_database
    try:
        _create_empty_database(man.database)
    except odoo.service.db.DatabaseExists:
        pass


def setup_company(man):
    country = man.company_spec.country_code

    with man.env() as env:
        # Post Init database with base module
        Country = env['res.country']
        company_country = Country.search(
            [
                ['code', 'ilike', country]
            ],
            limit=1
        )
        env.company.country_id = company_country.id


class DbApi(object):

    def __init__(self, manage, database):
        self.environment = manage.environment
        self.database = database
        self.config = manage.config
        self._entrypoint_loaded = False
        self.without_demo = True

    def mark_modules(self, modules, force=False):
        to_install = {mod for mod in modules}
        to_update = set()

        if not force:
            try:
                with self.env() as env:
                    IrModule = env['ir.module.module']
                    installed_modules = IrModule.search(
                        [
                            ['name', 'in', list(to_install)],
                            ['state', '=', 'installed'],
                        ],
                    )
                    for mod in installed_modules:
                        to_install.remove(mod.name)
                        to_update.add(mod.name)
            except Exception:
                pass

        for mod in to_install:
            self.config['init'][mod] = 1

        for mod in to_update:
            self.config['update'][mod] = 1

    def unmark_modules(self):
        for key in self.config['init'].keys():
            self.config['init'][key] = False

        for key in self.config['init'].keys():
            self.config['update'][key] = False

    def init(self, modules, country, language="en_US", without_demo=True):
        company = CompanySpec(country_code=country)

        self.company_spec = self.environment.manage.company_spec = company
        self.without_demo = self.environment.manage.without_demo = without_demo
        self.languages = self.environment.manage.languages = language

        self.environment.manage.initialize_odoo()

        self.install_modules_registry(
            ["base"],
            phase="Initializing database",
            event="initdb"
        )

        to_install = modules.copy()
        if 'base' in to_install:
            to_install.remove('base')

        self.install_modules_registry(
            modules,
            phase="Installing other modules",
            event="install_modules"
        )

    def default_entrypoints(self):
        if self._entrypoint_loaded:
            return

        self._entrypoint_loaded = True

        entrypoint("odoo_tools.manage.before_config")(before_config)
        entrypoint("odoo_tools.manage.after_config")(after_config)

        entrypoint("odoo_tools.manage.initialize_odoo")(initialize_odoo)
        entrypoint("odoo_tools.manage.before_initdb")(ensure_db)
        entrypoint("odoo_tools.manage.after_initdb")(setup_company)

    def install_modules(
        self,
        modules,
        force=False,
        phase="unknown",
        event="install_modules"
    ):
        self.environment.manage.initialize_odoo()
        self.install_modules_registry(
            modules,
            phase=phase,
            event=event
        )

    def install_modules_registry(
        self,
        modules,
        phase="unknown",
        force=False,
        event=None
    ):
        from odoo.service.server import preload_registries

        if event:
            execute_entrypoint(
                "odoo_tools.manage.before_{}".format(event),
                self
            )

        self.mark_modules(modules, force=force)

        # Creates a registry to be called to install
        rc = preload_registries([self.database])
        if rc != 0:
            _logger.error(
                "Failed installing or updating modules in phase %s",
                phase
            )
            raise InstallModulesError("Failed to install {}".format(modules))

        # Disable init for all modules
        self.unmark_modules()
        if event:
            execute_entrypoint(
                "odoo_tools.manage.after_{}".format(event),
                self
            )

    def uninstall_modules(
        self,
        modules
    ):
        self.environment.manage.initialize_odoo()

        with self.env() as env:
            IrModule = env['ir.module.module']

            installed_modules = IrModule.search(
                [
                    ['name', 'in', list(modules)],
                    ['state', '!=', 'uninstalled'],
                ],
            )

            installed_modules.button_immediate_uninstall()

    def export_translation_terms(self, languages, modules):
        """
        Export terms grouped by module
        """
        # TODO create a unified term exporter. The one from Odoo14 is
        # much better than previous versions. It could be a good start
        # as resources don't need to have odoo access.
        # Then implement a loader that can extract terms from fields
        # like Selection and translatable fields.
        #
        # Odoo14 implement selections through ir.models.fields.selection
        # while previous versions use the field.selection attribute.
        #
        # We don't really have to complicate ourselves as in odoo14
        # It's just a basic model being exported (need to guess proper xmlid)
        # While on earlier versions of odoo we can simply use a custom loader
        # for selection field.
        # In practice we could attempt to export translations using both format
        # and odoo should simply ignore older formats and everyone is happy.
        try:
            # Odoo 10, 11, 12, 13
            from odoo.tools import trans_generate
        except ImportError:
            # Odoo 14
            from odoo.tools.translate import TranslationModuleReader

            def trans_generate(language, modules, cr):

                reader = TranslationModuleReader(
                    cr,
                    modules=modules,
                    lang=language
                )
                return [
                    x
                    for x in reader
                ]

        def get_key(key_id):
            def wrap(value):
                return value[key_id]
            return wrap

        with self.env() as env:
            for language in languages:
                translations = trans_generate(
                    language,
                    modules,
                    env.cr
                )

                translations.sort()

                for module, terms in groupby(translations, key=get_key(0)):
                    yield language, module, terms

        _logger.info('translation file written successfully')

    @contextmanager
    def env(self, uid=None, ctx=None):
        import odoo

        if uid is None:
            uid = 1

        if ctx is None:
            ctx = {}

        odoo_registry = odoo.registry(self.database)

        # Post Init database with base module
        with closing(odoo_registry.cursor()) as cr:
            try:
                env = odoo.api.Environment(cr, uid, ctx)
                yield env
                env.cr.commit()
            except Exception:
                env.cr.rollback()



import string
import random
from .compat import Path
import configparser


def from_csv(value, custom_type=None):
    if not value:
        return []

    res = [
        val.strip() for val in value.split(',')
        if val.strip()
    ]

    if custom_type:
        res = custom_type(res)

    return res


def to_path_list(paths):
    return [
        Path(path)
        for path in paths
    ]


def is_subdir_of(path1, path2):
    try:
        path2.relative_to(path1)
    except ValueError:
        return False
    else:
        return True


def filter_excluded_paths(paths, excluded_paths):
    res_paths = []

    for vpath in paths:
        cur_path = Path(vpath)
        for ex_path in excluded_paths:
            if is_subdir_of(ex_path, cur_path):
                break
        else:
            res_paths.append(vpath)

    return res_paths


def convert_env_value(name, value):
    if value in ['True', 'False']:
        return value == 'True'

    return value


def random_string(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


class ConfigParser(configparser.ConfigParser):
    def set(self, section, option, value=None):
        if section not in self:
            self.add_section(section)
        return super(ConfigParser, self).set(section, option, value)



import click
import platform as pp


@click.group(
    help="Print the platform name to the output."
)
@click.pass_context
def platform(ctx):
    pass


@platform.command()
@click.pass_context
def arch(ctx):
    arch = pp.processor()

    if arch == "x86_64":
        docker_arch = "amd64"
    elif arch == "aarch64":
        docker_arch = "arm64"
    else:
        docker_arch = arch

    print(docker_arch, end="")



from odoo_tools.services.objects import ServiceManifests


inheritance_services = {
    'services': [
        {
            'name': 'prod',
            'inherit': 'dev',
            'env': {
                'a': '3',
                'd': '4'
            }
        },
        {
            'name': 'prod2',
            'inherit': 'dev',
            'addons': [],
            'env': {
                'a': '3',
                'd': '4'
            }
        },
        {
            'name': 'base',
            'addons': [
                {
                    'url': 'git@github.com:example/com.git',
                    'branch': 'main'
                }
            ],
            'env': {
                'a': 'b',
                'c': 'd'
            }
        },
        {
            'name': 'dev',
            'inherit': 'base',
            'env': {
                'a': '2',
                'b': '1'
            },
            'addons': [
                {
                    'url': 'git@github.com:example/com2.git',
                    'branch': 'staging'
                },
                {
                    'url': 'git@github.com:example/com.git',
                    'branch': 'staging'
                }
            ],
        }
    ]
}


def test_no_services():
    services = {
        'services': [
        ]
    }

    parsed_services = ServiceManifests.parse(services)

    assert len(parsed_services.services) == 0


def test_basic_service():
    services = {
        'services': [
            {
                'name': 'base',
                'env': {
                    'a': 'b',
                    'c': 'd'
                }
            }
        ]
    }

    parsed_services = ServiceManifests.parse(services)

    assert len(parsed_services.services) == 1
    assert parsed_services.services['base'].name == 'base'
    assert parsed_services.services['base'].env.to_dict() == {
        'a': 'b',
        'c': 'd'
    }

    dict_vals = {
        'services': {
            'base': {
                'name': 'base',
                'env': {
                    'a': 'b',
                    'c': 'd'
                },
                'addons': {}
            }
        }
    }

    assert parsed_services.to_dict() == dict_vals


def test_service_no_inheritance():
    manifest = ServiceManifests.parse(inheritance_services)

    assert len(manifest.services) == 4
    assert manifest.services['prod'].addons is None
    assert len(manifest.services['prod2'].addons) == 0
    assert len(manifest.services['dev'].addons) == 2
    assert len(manifest.services['base'].addons) == 1

    vals = {'a': '3', 'd': '4'}
    assert manifest.services['prod'].env.to_dict() == vals


def test_service_inheritance():
    manifest = ServiceManifests.parse(inheritance_services)

    assert len(manifest.services['prod'].resolved.addons) == 2
    assert len(manifest.services['dev'].resolved.addons) == 2
    assert len(manifest.services['base'].resolved.addons) == 1

    prod_resolved = manifest.services['prod'].resolved

    vals = {'a': '3', 'd': '4', 'b': '1', 'c': 'd'}
    assert prod_resolved.env.to_dict() == vals

    assert prod_resolved.addons['github_com_example_com'].branch == 'staging'



from odoo_tools.modules.search import find_addons_paths
from odoo_tools.api.objects import Manifest
from mock import patch, MagicMock


def test_find_paths_entrypoint(tmp_path):
    mocked_path = tmp_path / 'mocked'

    def mocked_iter_entrypoints():
        mock_entrypoint = MagicMock()

        mock_entrypoint.load.return_value = lambda: {mocked_path}

        return [mock_entrypoint]

    with patch('pkg_resources.iter_entry_points') as mock_pkg:
        mock_pkg.return_value = mocked_iter_entrypoints()

        options = MagicMock()
        options.include_odoo_entrypoints = True

        res = find_addons_paths(set(), options)

        assert res == set()

        manifest = Manifest(mocked_path / 'mod')
        manifest.save()

        res = find_addons_paths(set(), options)

        assert res == set([mocked_path])



from six import ensure_binary, ensure_str
from subprocess import PIPE
from .misc import run, cd
from tempfile import TemporaryDirectory
from urllib.parse import urlparse
import giturlparse
import os

import logging
from odoo_tools.compat import Path
from cryptography.fernet import Fernet
from contextlib import contextmanager

_logger = logging.getLogger(__name__)


@contextmanager
def ssh_command(addon, decrypt_key=None):
    old_command_exists = 'GIT_SSH_COMMAND' in os.environ
    old_command = os.environ.get('GIT_SSH_COMMAND')
    command_set = False

    if addon.private_key:
        key_data = ensure_binary(addon.private_key)

        # Decrypt key if possible
        if decrypt_key:
            fernet = Fernet(decrypt_key)
            key_data = fernet.decrypt(key_data)

        with TemporaryDirectory() as tempdir:
            key_dir = Path(tempdir)
            key_file = key_dir / 'key.pem'
            with key_file.open('wb') as keyfd:
                keyfd.write(key_data)

            os.chmod(str(key_file), 0o600)

            os.environ['GIT_SSH_COMMAND'] = 'ssh -i {}'.format(key_file)
            command_set = True

            yield
    else:
        yield

    if old_command_exists:
        os.environ['GIT_SSH_COMMAND'] = old_command
    elif command_set:
        del os.environ['GIT_SSH_COMMAND']


def fetch_addons(addon, output_directory, decrypt_key=None, credentials=None):
    if credentials is None:
        credentials = {}

    origin_url = addon.url

    parsed = giturlparse.parse(origin_url)

    host = parsed.host
    url_scheme = parsed.protocol

    if (
        (url_scheme == 'ssh' or addon.auth) and
        host in credentials
    ):
        credential = credentials.get(host)
        username = credential['username']
        password = credential['password']

        url = urlparse(parsed.url2https)
        url = url._replace(
            netloc="{}:{}@{}".format(
                username,
                password,
                host
            )
        ).geturl()
    else:
        url = origin_url

    _logger.info("Fetching %s", url)

    repo_path = Path.cwd() / output_directory / addon.repo_path

    repo_path.mkdir(exist_ok=True, parents=True)

    with cd(repo_path):
        run(['git', 'init'], check=False)
        run(['git', 'remote', 'add', 'origin', url], check=False)

        ref = addon.ref

        with ssh_command(addon, decrypt_key=decrypt_key):
            if ref:
                run(['git', 'fetch', 'origin', ref])
            else:
                run(['git', 'fetch', 'origin'])

        run(['git', 'checkout', 'FETCH_HEAD'])
        run(['git', 'remote', 'remove', 'origin'], check=False)

        commit_id = run(
            ['git', 'log', '--pretty=%H', '-1', 'FETCH_HEAD'],
            stdout=PIPE
        )

    return repo_path, {
        "url": url,
        "commit": ensure_str(commit_id).replace('\n', '')
    }


def checkout_repo(src, dest):
    env = dict(os.environ)
    env['GIT_WORK_TREE'] = str(dest)

    dest.mkdir(exist_ok=True, parents=True)

    _logger.info("Copying %s to %s", src, dest)

    with cd(src):
        run(
            ['git', 'checkout', '-f', 'FETCH_HEAD'],
            env=env
        )
        commit_id = run(
            ['git', 'log', '--pretty=%H', '-1'], stdout=PIPE
        )

    return commit_id



import os
from mock import patch
from odoo_tools.modules.search import Manifest
from odoo_tools.odoo import Environment
from odoo_tools.api.context import Context


def test_requirements(tmp_path):
    addons = tmp_path / 'addons'
    mod1 = Manifest(addons / 'mod1')
    mod1.set_attribute(['external_dependencies', 'python'], ['gitpython'])
    mod1.save()

    mod2 = Manifest(addons / 'mod2')
    mod2.set_attribute(['external_dependencies', 'python'], ['barcode'])
    mod2.save()

    requirement_file = addons / 'requirements.txt'

    with requirement_file.open('w') as fout:
        fout.write("""requests""")

    env = Environment()
    env.context.custom_paths.add(addons)

    requirements = env.modules.requirements()

    assert requirements == {'gitpython', 'barcode'}

    requirements_files1 = env.requirement_files()
    requirements_files2 = env.requirement_files(lookup_requirements=True)

    assert requirements_files1 == requirements_files2

    requirement_file2 = addons / 'mod2' / 'requirements.txt'

    with requirement_file2.open('w') as fout:
        fout.write("""gitlab""")

    requirements_files3 = env.requirement_files()
    requirements_files4 = env.requirement_files(lookup_requirements=True)

    assert requirements_files2 == requirements_files3
    assert requirements_files3 != requirements_files4

    assert requirements_files4 == {requirement_file, requirement_file2}


def test_requirements_no_modules(tmp_path):
    addons = tmp_path / 'addons'
    mod1 = Manifest(addons / 'mod1')
    mod1.save()

    mod2 = Manifest(addons / 'mod2')
    mod2.save()

    requirement_file = addons / 'requirements.txt'

    with requirement_file.open('w') as fout:
        fout.write("""requests""")

    env = Environment()
    env.context.custom_paths.add(addons)
    requirement_files = env.requirement_files()

    assert requirement_files == set([requirement_file])

    requirements = env.modules.requirements(extra_paths=requirement_files)

    assert requirements == {'requests'}


def test_disabled_modules(tmp_path):
    vals = dict(
        ODOO_DISABLED_MODULES="",
    )

    with patch.dict(os.environ, vals):
        context = Context.from_env()
        env = Environment(context)
        assert list(env.modules.disabled_modules()) == []



"""
Compatibility
=============

This module provides a common interface for functions that have
incompatibilities between python versions. This is an internal
module and you shouldn't import things here in your project as
they may be subject to change or disapear without notice.

.. note::
    This module is deprecated and will be refactored. Initially, its
    purpose was to provide a common interface for some libraries that
    had different interfaces between python2 and python3. As of now,
    odoo-tools doesn't attempt to work on python2 anymore.


"""
import sys
import signal
import shlex
# import re
import os
import subprocess
from pathlib import Path
# from configparser import ConfigParser, NoOptionError, NoSectionError
from importlib.util import find_spec

SIGSEGV = signal.SIGSEGV.value
quote = shlex.quote


def flush_streams():
    """
    Simple utility that flushes streams.

    When running in docker, it seems that it's required to manually flush
    the buffers to make logging more fluent. Otherwise, the logger in docker
    will not display anything until the buffers are filled enough. As logs
    in this library are relatively small. It usually never fill the logs fast
    enough.
    """
    sys.stdout.flush()
    sys.stderr.flush()


def log(message, *args, **kwargs):
    """
    Simple utility to log data to stdout.

    It calls ``flush_streams`` after writing to stdout.

    Parameters:
        message (str): The message to log
    """
    print(message, *args, **kwargs)
    flush_streams()


def module_path(module, raise_not_found=True):
    """
    A function that returns the path in which the module is located.

    Parameters:
        module (str): The name of the module to check.

    Returns:
        path (Path): the path of the module.
    """
    spec = find_spec(module)

    if not spec:
        if raise_not_found:
            raise ModuleNotFoundError(
                "Module {} cannot be found".format(module)
            )
        else:
            return None

    if spec.origin:
        return Path(spec.origin).parent
    else:
        for path in spec.submodule_search_locations:
            return Path(path)
        else:
            if raise_not_found:
                raise ModuleNotFoundError(
                    "Module {} cannot be found".format(module)
                )
            else:
                return None


def pipe(args):
    """
    Call the process with std(in,out,err)

    Parameters:
        args (List<str>): A list of parameters to be passed to Popen.

    Returns:
        returncode (int): The returncode of the program
    """
    log("Executing external command %s" % " ".join(args))

    env = os.environ.copy()
    env['DEBIAN_FRONTEND'] = 'noninteractive'

    process = subprocess.Popen(
        args,
        stdin=sys.stdin,
        stdout=sys.stdout,
        stderr=sys.stderr,
        env=env
    )

    process.wait()

    log(
        (
            "External command execution completed with returncode(%s)"
        ) % process.returncode
    )

    if process.returncode == -SIGSEGV:
        log("PIPE call segfaulted")
        log("Failed to execute %s" % args)

    return process.returncode



import six
from six import ensure_binary, ensure_text
import shutil
import logging
from ast import literal_eval
import tempfile
from zipfile import ZipFile

from ..compat import Path
from ..modules.render import render_description_str
from ..modules.translate import PoFileWriter, PoFileReader
from ..exceptions import ArgumentError

_logger = logging.getLogger(__name__)


class CompanySpec(object):
    """
    This class represent a company object.

    This object can be used along database initialization.
    When creating a new database, it is possible to configure the
    main company before installing more modules. One example where
    the company is particularly important is when you attempt to
    install the account module.

    Without a valid company, Odoo will try to install a default
    l10n module for accounting for a locale that you would likely
    not use.

    When a company object is provided to the database initialization,
    it will start by initializing the database with the base module.

    Then when the base module is installed, it will update some objects
    like the `res.company` and then it will install other modules that
    were requested to be installed for the database initialization.

    Attributes:
        country_code (str): Country code based on the Alpha-2 of the
            `ISO 3166-1 <ISO>`_ code.
    .. _ISO: https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes
    """
    def __init__(self, country_code):
        self.country_code = country_code


def try_parse_manifest(data):
    """
    Attempt parsing a manifest.

    It used the literal_eval native function to attempt parsing
    a manifest.

    Args:
        data (str): Content of the manifest to load

    Returns:
        dict: The parsed data of the manifest.
    """
    return literal_eval(data)


def try_compile_manifest(data):
    """
    Attempt compiling the manifest.

    In some specific cases, if the literal eval doesn't work. It can
    provide more information to compile the manifest and evaluate its
    content.

    As manifest file are python modules, it is technically possible
    to import or use python code directly within the manifest file.

    In that case a simple literal_eval wouldn't be enough. That said,
    this evaluator should never be used.

    Args:
        data (str): Content of the manifest to load

    Returns:
        dict: The parsed data of the manifest.
    """
    code = compile(
        data,
        '__manifest__.py',
        'eval'
    )
    return eval(code, {}, {})


def get_translation_filename(language, module):
    """
    Get the filename for translation file.

    When no language is provided to translate a module,
    the name of the file will the name of the module as
    a pot file.

    For example, exporting the pot file for module bus would be
    exported by calling:

    .. code-block:: python

        get_translation_filename(None, 'bus')
        >> 'bus.pot'

    When exporting the translation file for fr_FR language for
    the module bus. It will return the family of the language
    as the langage and dialect are the same.

    .. code-block:: python

        get_translation_filename('fr_FR', 'bus')
        >> 'fr.po'

    If the language and dialect are different it returns the language
    locale as is.

    .. code-block:: python

        get_translation_filename('fr_CA', 'bus')
        >> 'fr_CA.po'

    Args:
        language (str|None): locale to export

        module (str): Name of the module to export

    Returns:
        str: The filename of the translation file based
            on the arguments.
    """
    if language and '_' in language:
        family, dialect = language.lower().split('_')
        if family == dialect:
            language = family

    filename = (
        '{}.po'.format(language)
        if language
        else '{}.pot'.format(module)
    )
    return filename


class Manifest(object):
    """
    Manifest Object

    The manifest object contains data stored in odoo/openerp manifests.

    Attributes:
        _attrs (dict): Data of the manifest file. You can see the whole set
            of attributes in the `odoo documentation <odoo>`_.

        path (Path): The path in which the manifest got loaded.

        _manifest_file: The path of the manifest file of this manifest
            if it was loaded from a path. If the path of the manifest
            wasn't provided, it defaults to the filename `__manifest__.py`
            in the path of the module.

        technical_name: This attribute is stored into the _attrs but is defined
            as the module name. The name attribute in a manifest is just a name
            that will then get loaded in odoo. The technical name is the unique
            identifier for the module path in the addons paths.

    .. _odoo: https://www.odoo.com/documentation/master/developer/reference/backend/module.html
    """  # noqa

    properties = {
        "_attrs",
        "path",
        "_manifest_file"
    }

    defaults = {
        "installable": True,
        "application": False,
        "depends": list,
        "demo": list,
        "data": list,
        "version": "0.0.0",
        "external_dependencies": dict
    }

    def __init__(self, path, attrs=None, manifest_file=None):
        """
        Initialize manifest objects.

        Args:
            path (Path): Path in which the module is located

            attrs (dict): Data to set in the manifest.

            manifest_file (Path): Location of the manifest itself if provided.
        """
        if attrs is None:
            attrs = {}

        self._attrs = attrs
        self.path = Path(path)

        if 'technical_name' not in attrs:
            attrs['technical_name'] = self.path.name

        self._manifest_file = manifest_file or (self.path / '__manifest__.py')

        self.set_defaults()

    def set_defaults(self):
        """
        Set some defaults when initializing an object.

        It uses the values as defined in `Manifest.defaults`. If a value
        is already present in `self._attrs`, it will not update the value
        from the defaults.
        """
        for key, value in Manifest.defaults.items():
            if key not in self._attrs:
                if callable(value):
                    self._attrs[key] = value()
                else:
                    self._attrs[key] = value

    def __getattr__(self, name, default=None):
        """
        Get an attribute from `self._attrs`.
        """
        return self._attrs.get(name, default)

    def __setattr__(self, name, value):
        """
        Set an attribute into the manifest. If it is a protected
        attribute, it will use the super method, otherwise it will
        set the custom attributes in `self._attrs`.

        The reason why some attributes are separate is that you
        wouldn't want to save some attributes to the manifest. For
        example, the `path` of the module doesn't have to get saved.
        """
        if (
            name in Manifest.properties
            # name in Manifest.defaults.keys()
        ):
            super(Manifest, self).__setattr__(name, value)
        else:
            self._attrs[name] = value

    def __lt__(self, other):
        """
        Comparison function using the name property.

        It is required to make Manifest sortable in a list by name.
        Sorting modules lists can be required in order to give
        consistent results.

        For example, a simple text diff would fail if the module set
        was returned in random order. When sorting the modules in a list,
        the output will always be the same if both sets are the same.

        Args:
            other (Manifest): The other manifest to be compared against.

        Returns:
            bool: if the name is bigger than the other name.
        """
        return self.path.name < other.path.name

    def __eq__(self, other):
        """
        Compare a manifest against something else.

        This method will compare an other object with the path name of
        the current manifest. The other will be converted to a path name
        if possible and compared. If it's not possible to compare it to
        a path name, then it is considered as unequal.

        Args:
            other (Manifest|str|Path,Any): the other to be checked against.
        """
        if isinstance(other, Manifest):
            return self.path == other.path
        elif isinstance(other, str):
            return self.path.name == Path(other).name
        elif isinstance(other, Path):
            return self.path.name == other.name
        else:
            return False

    def __contains__(self, value):
        """
        Check if a property is defined in properties of the
        manifest.

        Example:

        .. code-block: python

            'technical_name' in self
            >> True

        Args:
            value (str): Name of a property

        Returns:
            bool: If the property is present in the manifest.
        """
        return value in self._attrs

    def __hash__(self):
        """
        Makes the manifest hashable based on its name.

        In practice two modules in different path would get
        the same hash. The reason for this is that odoo doesn't
        handle having 2 different modules with the same path name.

        When importing a module it could pick one or the other module
        based on how `odoo.addons` searches for modules within the
        defined paths.

        Obviously, you should check if you have 2 modules with the
        same technical name.
        """
        return hash(self.path.name)

    def __str__(self):
        """
        Serialize the str interpretation of a manifest.
        """
        return str(self.path)

    def __repr__(self):
        """
        Python representation of a manifest with all its properties.
        """
        return "Manifest({path}, {attrs})".format(
            path=repr(self.path),
            attrs=repr(self._attrs)
        )

    def values(self):
        """
        Returns the manifest's properties.
        """
        return self._attrs

    def set_attribute(self, attribute, value):
        """
        Set attribute to a value.

        This is a shortcut method to set attributes in a nested dict.

        .. code-block:: python

            manifest.set_attribute(
                ['external_dependencies', 'python'], ['request']
            )
            assert manifest.external_dependencies['python'] == ['request']

        Args:
            attribute (list(str)): List of attribute key to set

            value (any): Value to set to the attribute path
        """
        if len(attribute) == 0:
            raise ArgumentError(
                'The attribute must have at least one attribute name.'
            )
        cur_vals = self._attrs

        last_attribute = attribute[-1]

        for attr in attribute[:-1]:
            if attr not in cur_vals:
                cur_vals[attr] = {}
            cur_vals = cur_vals[attr]
        else:
            cur_vals[last_attribute] = value

    def static_assets(self):
        """
        Returns static asset files.

        This lookup all files in the static folder of the current
        module/manifest. It will return all files without exceptions.

        This can be mainly used to lookup for static assets of specific
        modules. For example, you'd want to upload all static assets
        somewhere else or you'd want to preprocess static assets before
        deployment.

        Returns:

            list(Path): A list of files static subdirectory
        """
        addons_path = self.path.parent

        def recurse_search(cur_dir):
            files = []
            for file in cur_dir.iterdir():
                if file.is_dir():
                    files += recurse_search(file)
                else:
                    static_path = file.relative_to(addons_path)
                    files.append((static_path, file))
            return files

        if (self.path / "static").exists():
            return recurse_search(self.path / "static")
        else:
            return []

    @classmethod
    def from_path(klass, manifest, render_description=False):
        """
        Loads the manifest from a given path.

        It automatically define the module name to the path name of the
        folder in which the manifest is located to those attributes:

        - name
        - technical_name

        A custom attribute `description_html` can be generated to store
        the rendered html description.

        Args:
            manifest (Path): The path of the manifest or module.
            render_description (bool): Set to True if you need to render
              the description.

        Returns:
            Manifest: The manifest that was loaded.
        """
        if not manifest.name.endswith('py'):
            module_path = manifest
            odoo_manifest = module_path / '__manifest__.py'
            erp_manifest = module_path / '__openerp__.py'

            manifest = None
            if odoo_manifest.exists():
                manifest = odoo_manifest
            if not manifest and erp_manifest.exists():
                manifest = erp_manifest
        else:
            module_path = manifest.parent

        with manifest.open('rb') as fin:
            manifest_data = ensure_text(fin.read() or "")
            manifest_data = manifest_data.replace('\ufeff', '')

        parsers = [
            try_parse_manifest,
            try_compile_manifest,
            # try_parse_manifest_encoded,
            # try_compile_manifest_encoded
        ]

        last_error = None
        for parser in parsers:
            try:
                data = parser(manifest_data)
                break
            except Exception as exc:
                last_error = exc
        else:
            data = {}
            if last_error is not None:
                _logger.error(
                    "Cannot parse manifest: {}".format(
                        manifest,
                    ),
                )
                raise last_error

        if 'name' not in data:
            data['name'] = module_path.name

        data['technical_name'] = module_path.name

        if render_description:
            data['description_html'] = render_description_str(
                module_path,
                data.get('description', '')
            )

        man = Manifest(module_path, attrs=data, manifest_file=manifest)

        return man

    def save(self):
        """
        Saves changes made to the manifest in python.

        All changes will be dumped into the original manifest
        file or in __manifest__.py.
        """
        obj = {}
        for key, value in self._attrs.items():
            if key in ['description_html', 'technical_name']:
                continue

            if isinstance(value, Path):
                obj[key] = str(value)
            else:
                obj[key] = value

        if not self._manifest_file.parent.exists():
            self._manifest_file.parent.mkdir(parents=True)

        with self._manifest_file.open('w') as fout:
            fout.write(repr(obj))

    def requirements(self, package_map=None):
        """
        Return a list of packages required by module.
        """
        packages = set()

        for package in self.external_dependencies.get('python', []):
            package_lower = package.lower()

            if package_map and package_lower in package_map:
                package_name = package_map[package.lower()]
                if package_name:
                    packages.add(package_name)
            else:
                packages.add(package)

        return packages

    def disable(self):
        """
        Set the manifest as uninstallable.
        """
        self.installable = False

    def remove(self):
        """
        Remove the module from the file system.

        It will attempt to remove the folder in which the
        manifest is located.

        If the directory doesn't exist, it does nothing.
        If the directory does exist. It will try to remove
        all files located in the module folder recursively.
        """
        if self.path.exists():
            shutil.rmtree(self.path, ignore_errors=True)

    def files(self):
        """
        Yields all file located in the module's folder.

        Iterating over this iterator yields all file in the module's
        folder except pyc files or potentially other files that aren't
        considered as "sources". It will also yield files located in
        the static folder.
        """
        for file in self.path.glob('**/*'):
            if file.is_dir():
                continue
            if file.name.endswith('.pyc'):
                continue

            yield file

    def package(self):
        """
        Returns a file handle to a zipfile of the packaged module.

        This methods will generate a zip file based on the result of
        a call to `self.files()`.

        The zipfile will get destroyed once it is closed or garbage
        collected.

        Returns:
            File: A file handle containing the zipped module.
        """
        outfile = tempfile.NamedTemporaryFile()

        root_folder = self.path.parent
        zipfile = ZipFile(outfile, 'w')

        for file in self.files():
            zip_filename = file.relative_to(root_folder)

            with file.open('rb') as fin:
                with zipfile.open(str(zip_filename), mode='w') as fout:
                    fout.write(fin.read())

        return outfile

    def export_translations(self, db, languages):
        """
        Exports translation in the corresponding module's path.

        It will store all potential translations into the i18n folder
        of the corresponding module.

        Args:
            db (DbApi): the api used to access the database.

            languages (list(str)): list of locales to export.

        Returns:
            Nothing
        """

        translations = db.export_translation_terms(
            languages,
            [self.technical_name]
        )

        for language, module, rows in translations:
            filename = get_translation_filename(language, module)
            trans_path = self.path / 'i18n' / filename
            trans_path.parent.mkdir(parents=True, exist_ok=True)

            _logger.info(
                "Exporting translation %s of module %s to %s",
                language,
                module,
                trans_path
            )

            if trans_path.exists():
                origin_po_file = PoFileReader(trans_path.open('rb'))
            else:
                with trans_path.open("w+") as buffer:
                    buffer.write("")

                with trans_path.open('rb') as buffer:
                    origin_po_file = PoFileReader(buffer)

            with trans_path.open('wb+') as buffer:
                po_writer = PoFileWriter(
                    buffer,
                    language,
                    pofile=origin_po_file.pofile
                )

                po_writer.add_entries(rows)

                po_writer.write()



"""
Odoo Translation
================

This module has a collection of tools that can be used
to extract translations from Odoo. It needs a working instance of
Odoo to be able to import/export translations out of Odoo.

It can do some of the translation file generation on its own but
in order to prefill some already translated terms, it has to be able
to load them into odoo first.
"""
import re
import polib
import logging
from ..compat import Path
from datetime import datetime

_logger = logging.getLogger()


def get_modules(entry):
    try:
        match = re.match(r"(module[s]?): (.+)", entry.comment)
        _, modules = match.groups()
        modules = modules.split(',')
    except Exception:
        modules = []

    modules_lst = []
    for module in modules:
        module = module.strip()
        if module not in modules_lst:
            modules_lst.append(module)
    modules = tuple(modules_lst)
    return modules


def get_comments(entry):

    comments = [
        c
        for c in entry.comment.split('\n')
        if not c.startswith('module:')
        if not c.startswith('modules:')
    ]

    return comments


class PoFileReader(object):
    """ Iterate over po file to return Odoo translation entries """
    def __init__(self, source, options=None):

        if options is None:
            options = {
                "strict": False,
                "read_pot": False
            }

        def get_pot_path(source_name):
            # when fileobj is a TemporaryFile, its name is an inter in P3, a
            # string in P2
            if isinstance(source_name, str) and source_name.endswith('.po'):
                # Normally the path looks like /path/to/xxx/i18n/lang.po
                # and we try to find the corresponding
                # /path/to/xxx/i18n/xxx.pot file.
                # (Sometimes we have 'i18n_extra' instead of just 'i18n')
                path = Path(source_name)
                filename = path.parent.parent.name + '.pot'
                pot_path = path.with_name(filename)
                return pot_path.exists() and str(pot_path) or False
            return False

        # polib accepts a path or the file content as a string, not a fileobj
        if isinstance(source, str):
            self.pofile = polib.pofile(source)
            pot_path = get_pot_path(source)
        elif isinstance(source, polib.POFile):
            self.pofile = source
        else:
            # either a BufferedIOBase or result from NamedTemporaryFile
            self.pofile = polib.pofile(source.read().decode())
            pot_path = get_pot_path(source.name)

        if options.get('read_pot') and pot_path:
            # Make a reader for the POT file
            # (Because the POT comments are correct on GitHub but the
            # PO comments tends to be outdated. See LP bug 933496.)
            self.pofile.merge(polib.pofile(pot_path))

    def __iter__(self):
        for entry in self.pofile:
            if entry.obsolete:
                continue

            modules = get_modules(entry)
            comments = get_comments(entry)

            source = entry.msgid
            translation = entry.msgstr
            found_code_occurrence = False
            for occurrence, line_number in entry.occurrences:
                match = re.match(
                    r'(model|model_terms):([\w.]+),([\w]+):(\w+)\.([^ ]+)',
                    occurrence
                )
                if match:
                    type, model_name, field_name, module, xmlid = match.groups() # noqa

                    yield {
                        'type': type,
                        'imd_model': model_name,
                        'name': model_name+','+field_name,
                        'imd_name': xmlid,
                        'res_id': None,
                        'src': source,
                        'value': translation,
                        'comments': "\n".join(comments),
                        'module': modules,
                    }
                    continue

                match = re.match(r'(code):([\w/.]+)', occurrence)
                if match:
                    type, name = match.groups()
                    if found_code_occurrence:
                        # unicity constrain on code translation
                        continue
                    found_code_occurrence = True
                    yield {
                        'type': type,
                        'name': name,
                        'src': source,
                        'value': translation,
                        'comments': comments,
                        'res_id': int(line_number),
                        'module': modules,
                    }
                    continue

                match = re.match(r'(selection):([\w.]+),([\w]+)', occurrence)
                if match:
                    if self.options.get('strict'):
                        _logger.info(
                            "Skipped deprecated occurrence %s", occurrence
                        )
                        continue

                    type, model_name, name = match.groups()

                    yield {
                        'type': type,
                        'name': name,
                        'imd_model': model_name,
                        'res_id': int(line_number),
                        'src': source,
                        'value': translation,
                        'comments': comments,
                        'module': modules
                    }

                    continue

                match = re.match(
                    r'(sql_constraint|constraint):([\w.]+)', occurrence
                )
                if match:
                    _logger.info(
                        "Skipped deprecated occurrence %s", occurrence
                    )
                    continue
                _logger.error(
                    "malformed po file: unknown occurrence: %s", occurrence
                )


class PoFileWriter(object):
    """ Iterate over po file to return Odoo translation entries """
    def __init__(self, target, lang, pofile=None):

        self.buffer = target
        self.lang = lang
        if pofile is not None:
            self.po = pofile
        else:
            self.po = polib.POFile()

    def merge(self, pofile):
        self.po.merge(pofile)

    def write_rows(self, rows):
        # we now group the translations by source. That means one translation
        # per source.
        grouped_rows = {}
        modules = set([])

        for module, type, name, res_id, src, trad, comments in rows:
            row = grouped_rows.setdefault(src, {})
            row.setdefault('modules', set()).add(module)
            if not row.get('translation') and trad != src:
                row['translation'] = trad
            row.setdefault('tnrs', []).append((type, name, res_id))
            row.setdefault('comments', set()).update(comments)
            modules.add(module)

        for src, row in sorted(grouped_rows.items()):
            if not self.lang:
                # translation template, so no translation value
                row['translation'] = ''
            elif not row.get('translation'):
                row['translation'] = ''

            self.add_entry(
                row['modules'],
                sorted(row['tnrs']),
                src,
                row['translation'],
                row['comments']
            )

        self.write()

    def generate_header(self):
        import odoo.release as release

        modules = set([])

        self.po.header = (
            "Translation of %s.\n"
            "This file contains the translation of the following modules:"
            "\n%s"
        ) % (
            release.description, ''.join("\t* %s\n" % m for m in modules)
        )

    def generate_metadata(self):
        import odoo.release as release

        now = datetime.utcnow().strftime('%Y-%m-%d %H:%M+0000')
        self.po.metadata = {
            'Project-Id-Version': "%s %s" % (
                release.description, release.version
            ),
            'Report-Msgid-Bugs-To': '',
            'POT-Creation-Date': now,
            'PO-Revision-Date': now,
            'Last-Translator': '',
            'Language-Team': '',
            'MIME-Version': '1.0',
            'Content-Type': 'text/plain; charset=UTF-8',
            'Content-Transfer-Encoding': '',
            'Plural-Forms': '',
        }

    def write(self):
        if not self.po.header:
            self.generate_header()
        if not self.po.metadata:
            self.generate_metadata()

        self.buffer.write(str(self.po).encode())

    def add_entry(self, modules, tnrs, source, trad, comments=None):
        entry = self.po.find(st=source)

        if not entry:
            entry = polib.POEntry(
                msgid=source,
                msgstr=trad,
            )
            self.po.append(entry)

        entry_modules = list(get_modules(entry))
        for module in modules:
            if module not in entry_modules:
                entry_modules.append(module)

        modules = entry_modules

        plural = len(modules) > 1 and 's' or ''
        entry.comment = "module%s: %s" % (plural, ', '.join(modules))

        if comments:
            entry_comments = get_comments(entry)
            for comment in comments:
                if comment not in entry_comments:
                    entry_comments.append(comment)
            comments = entry_comments

            entry.comment += "\n" + "\n".join(comments)

        code = False
        for typy, name, res_id in tnrs:
            if typy == 'code':
                code = True
            #     res_id = 0
            new_occurrence = None
            if isinstance(res_id, int) or res_id.isdigit():
                # second term of occurrence must be a digit
                # occurrence line at 0 are discarded when rendered to string
                new_occurrence = (u"%s:%s" % (typy, name), str(res_id))
            else:
                new_occurrence = (u"%s:%s:%s" % (typy, name, res_id), '')

            if new_occurrence not in entry.occurrences:
                entry.occurrences.append(new_occurrence)
        if code and "python-format" not in entry.flags:
            entry.flags.append("python-format")

        entry.flags = list(set(entry.flags))

    def add_entries(self, entries):
        for module, type, name, res_id, source, value, comments in entries:
            self.add_entry(
                modules=[module],
                tnrs=[
                    (type, name, res_id)
                ],
                source=source,
                trad=value,
                comments=comments
            )



import logging

from ..compat import pipe

_logger = logging.getLogger(__name__)


def install_apt_packages(odoo_env):
    """
    Install debian dependencies.
    """
    package_list = odoo_env.manage.packages()

    if len(package_list) > 0:
        _logger.info(
            "Installing %s",
            package_list
        )
        ret = pipe(['apt-get', 'update'])

        # Something went wrong, stop the service as it's failing
        if ret != 0:
            raise SystemError("Apt Update failed")

        base_install_params = ['apt-get', 'install', '-y']

        if not odoo_env.context.apt_install_recommends:
            base_install_params.append('--no-install-recommends')

        ret = pipe(
            base_install_params,
            list(package_list)
        )

        # Something went wrong, stop the service as it's failing
        if ret != 0:
            raise SystemError("Failed to install packages")


def fix_access_rights(odoo_env):
    # Change to some python alternative of chown/chmod
    # provide a way to configure user:group names
    if odoo_env.context.reset_access_rights == 'TRUE':
        pipe(["chown", "-R", "odoo:odoo", "/var/lib/odoo"])
        pipe(["chown", "-R", "odoo:odoo", "/etc/odoo"])


def remove_sudo():
    # TODO maybe add a file in sudoers.d and simply ensure it is removed
    return pipe(["sed", "-i", "/odoo/d", "/etc/sudoers"])



import click


@click.group()
def config():
    pass


@config.command("set")
@click.option('-s', '--section', default='options')
@click.argument('key')
@click.argument('value')
@click.pass_context
def set_value(ctx, section, key, value):
    env = ctx.obj['env']

    with env.config() as config:
        config.set(section, key, value)


@config.command("get")
@click.option('-s', '--section', default='options')
@click.argument('key')
@click.pass_context
def get_value(ctx, section, key):
    env = ctx.obj['env']

    with env.config() as config:
        print(config.get(section, key))


@config.command("ls")
@click.pass_context
def list_values(ctx):
    env = ctx.obj['env']

    with env.config() as config:
        for key, value in config._sections.items():
            for key2, value in value.items():
                print("{}.{} = {}".format(key, key2, value))



import toml
import json
import pytest
from unittest.mock import patch, MagicMock
from click.testing import CliRunner

import os
from odoo_tools.cli.odot import command
# from odoo_tools.cli import fetch_addons, copy_addons


@pytest.fixture
def runner():
    return CliRunner()


def test_command():
    with pytest.raises(SystemExit):
        command()


def test_shell(runner):
    with patch('ptpython.repl.embed') as mocked_func:
        mocked_func.return_value = 10
        result = runner.invoke(command, ["shell", "testdb"])
        msg = "Odoo doesn't seem to be installed. Check your path\n"
        assert not result.exception
        assert result.output == msg


def test_list_modules(runner, tmp_path):
    result = runner.invoke(command, ['module', 'ls'])
    assert not result.exception
    assert result.output == ""


def test_list_modules2(runner, tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True)

    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(command, ['module', 'ls', '--only-name'])

    assert not result.exception
    modules = set(x for x in result.output.split('\n') if x)
    assert modules == {'a', 'b', 'c', 'd'}


def test_list_modules_installable(runner, tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True, installable=False)

    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'ls', '--only-name', '--installable']
        )

    assert not result.exception
    modules = set(x for x in result.output.split('\n') if x)
    assert modules == {'a', 'b', 'c'}

    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'ls', '--only-name', '--non-installable']
        )

    assert not result.exception
    modules = set(x for x in result.output.split('\n') if x)
    assert modules == {'d'}


def test_show_modules(runner, tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True, installable=False)

    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'show', 'd']
        )

    assert isinstance(result.exception, SystemExit)

    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'show', 'a']
        )

    expected_data = {
        'application': False,
        'data': [],
        'demo': [],
        'depends': ['web'],
        'description': 'a',
        'external_dependencies': {},
        'installable': True,
        'name': 'a',
        'technical_name': 'a',
        'version': '0.0.0',
    }

    data = json.loads(result.output)
    assert data == expected_data


def generate_odoo_dir(tmp_path):
    odoo_dir = tmp_path / "odoo"
    odoo_dir.mkdir()
    addons_dir = odoo_dir / "addons"
    addons_dir.mkdir()

    return odoo_dir, addons_dir


def generate_addons(addons_dir, modules, **kw):
    for module in modules:
        mod_dir = addons_dir / module
        mod_dir.mkdir()

        manifest = mod_dir / '__manifest__.py'

        with manifest.open('w') as man:
            module_dict = {
                "description": module,
                "depends": ["web"]
            }

            for key, value in kw.items():
                module_dict[key] = value

            man.write(repr(module_dict))


def test_requirements(runner, tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True, installable=False)
    generate_addons(
        addons_dir, ['e'], external_dependencies={'python': ['requests']}
    )
    generate_addons(
        addons_dir, ['f'], external_dependencies={'python': ['Pillow']}
    )
    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'requirements', '--sort']
        )

    assert result.output == "pillow\nrequests\n"


def test_deps(runner, tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'], depends=[])
    generate_addons(addons_dir, ['d'], depends=['a'])
    generate_addons(addons_dir, ['e'], depends=['b'])
    generate_addons(addons_dir, ['z'], depends=['c'])
    generate_addons(addons_dir, ['f'], depends=['e', 'd', 'z'])
    generate_addons(addons_dir, ['h'], depends=['b'], auto_install=True)
    generate_addons(addons_dir, ['g'], depends=['f', 'h'], auto_install=True)
    with patch.dict(os.environ, {'ODOO_BASE_PATH': str(odoo_dir)}):
        result = runner.invoke(
            command,
            ['module', 'deps', '-m', 'f', '--auto', '--only-name', '--csv']
        )

    assert result.output == "a,b,c,d,e,h,z,g"


def test_service(runner, tmp_path):
    services = {
        "services": [
            {
                "name": "base",
                "addons": [
                    {
                        "url": "https://github.com/oca/web.git",
                    },
                    {
                        "url": "self",
                        "branch": "12.0-tests"
                    }
                ]
            },
            {
                "name": "test",
                "inherit": "base",
                "odoo": {
                    "version": "12.0",
                    "repo": {
                        "url": "https://github.com/odoo/odoo.git",
                    }
                },
                "addons": [
                    {
                        "url": "https://github.com/oca/knowledge.git"
                    }
                ]
            }
        ]
    }

    manifest_toml = tmp_path / 'services.toml'

    with manifest_toml.open('w') as fout:
        toml.dump(services, fout)

    result = runner.invoke(
        command,
        [
            'service',
            'show',
            '--url', 'https://github.com/llacroix/addons.git',
            str(manifest_toml),
            "test"
        ],
    )

    assert result.exception is None

    output = json.loads(result.output)

    out_data = {
        "name": "test",
        "odoo": {
            "version": "12.0",
            "repo": {
                "url": "https://github.com/odoo/odoo.git",
                'branch': None,
                'commit': None,
                'private_key': None,
                'ref': '12.0',
            },
            "options": {}
        },
        "env": {},
        "labels": {},
        "addons": {
            "github_com_oca_web": {
                "url": "https://github.com/oca/web.git",
                'branch': None,
                'commit': None,
                'private_key': None,
                'ref': '12.0',
            },
            "github_com_oca_knowledge": {
                "url": "https://github.com/oca/knowledge.git",
                'branch': None,
                'commit': None,
                'private_key': None,
                'ref': '12.0',
            },
            "self": {
                "url": "https://github.com/llacroix/addons.git",
                'branch': "12.0-tests",
                'commit': None,
                'private_key': None,
                'ref': '12.0-tests',
            }
        }
    }

    assert output == out_data


def test_odot_platform(runner, tmp_path):
    result = runner.invoke(
        command,
        [
            'platform',
            'arch'
        ],
    )

    assert result.output == "amd64"


def test_config(env, tmp_path, runner):
    odoo_rc_path = tmp_path / 'odoo.cfg'

    envs = {
        'ODOO_RC': str(odoo_rc_path)
    }

    runner.env = envs

    result = runner.invoke(
        command,
        [
            'config',
            'get',
            'data_dir'
        ],
        env=envs,
    )

    assert result.output == ""

    result = runner.invoke(
        command,
        [
            'config',
            'set',
            'data_dir',
            'value'
        ],
        env=envs,
    )

    assert result.output == ""

    result = runner.invoke(
        command,
        [
            'config',
            'get',
            'data_dir',
        ],
        env=envs,
    )

    assert result.output == "value\n"

    result = runner.invoke(
        command,
        [
            'config',
            'ls',
        ]
    )
    assert result.output == "options.data_dir = value\n"



import pkg_resources
from collections import defaultdict

custom_entrypoints = defaultdict(list)


def execute_entrypoint(name, env):
    for ep in pkg_resources.iter_entry_points(group=name):
        functor = ep.load()
        functor(env)

    for ep in custom_entrypoints.get(name, []):
        ep(env)


def entrypoint(name):
    def wrap(func):
        if func not in custom_entrypoints[name]:
            custom_entrypoints[name].append(func)
        return func
    return wrap



import toml
import pytest
from odoo_tools.odoo import Environment
from odoo_tools.compat import Path
from odoo_tools.exceptions import OdooNotInstalled, NoConfigError
from odoo_tools.api.objects import Manifest

from tests.utils import (
    generate_addons,
    generate_odoo_dir,
    generate_addons_full
)


def test_check_odoo(tmp_path):
    env = Environment()

    with pytest.raises(OdooNotInstalled):
        env.check_odoo()

    assert env.package_map() == {}


def test_package_map_exists(env, tmp_path):
    package_map_file = tmp_path / 'package.toml'

    with package_map_file.open('w') as fout:
        toml.dump({'xxx': 'yyy'}, fout)

    env.context.package_map_file = package_map_file
    env.context.odoo_version = "15.0"

    new_map = {
        "xxx": "yyy",
        "ldap": "python-ldap",
    }

    assert env.package_map() == new_map


def test_package_map_not_exists(env, tmp_path):
    package_map_file = tmp_path / 'package.toml'

    env.context.package_map_file = package_map_file
    env.context.odoo_version = "10.0"

    new_map = {
    }

    assert env.package_map() == new_map


def test_env_config(tmp_path):
    env = Environment()

    env.context.odoo_rc = tmp_path / 'odoo.cfg'

    with env.config():
        env.set_config('xmlrcp_interface', 'localhost')
        env.set_config('xmlrcp_port', '8069')

    with env.config():
        assert env.get_config('xmlrcp_interface') == 'localhost'

    with pytest.raises(NoConfigError):
        env.get_config('addons_paths')

    with pytest.raises(NoConfigError):
        env.set_config('addons_paths', '1')


def test_env_config_addons_path(tmp_path):
    env = Environment()

    env.context.odoo_rc = tmp_path / 'odoo.cfg'

    with env.config():
        env.set_config('addons_paths', ",".join([str(tmp_path)]))

    paths = env.addons_paths()
    assert paths == {tmp_path}

    env.context.force_addons_lookup = True
    assert env.addons_paths() == set()


def test_invalid_env(tmp_path):
    env = Environment()

    assert env.modules.list() == set()

    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)

    env.context.odoo_base_path = str(odoo_dir)

    assert len(env.modules.list()) == 0

    generate_addons(addons_dir, ['a', 'b', 'c'])

    assert len(env.modules.list()) == 3
    assert set(env.modules.server_wide_modules()) == {'web', 'base'}


def test_server_wide_modules(tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True)

    env = Environment()
    env.context.odoo_base_path = str(odoo_dir)

    assert len(env.modules.list()) == 4
    assert set(env.modules.server_wide_modules()) == {'web', 'base', 'd'}


def test_addons_paths(tmp_path):
    odoo_dir, addons_dir = generate_odoo_dir(tmp_path)
    generate_addons(addons_dir, ['a', 'b', 'c'])
    generate_addons(addons_dir, ['d'], server_wide=True)

    env = Environment()
    env.context.odoo_base_path = str(odoo_dir)

    assert len(env.modules.list()) == 4
    assert len(env.addons_paths()) == 1

    custom_addons_dir = addons_dir / 'addons2'
    custom_addons_dir.mkdir()

    generate_addons(custom_addons_dir, ['f', 'g', 'h'])

    assert len(env.modules.list(reload=True)) == 7
    assert len(env.addons_paths()) == 2

    custom_dir = tmp_path / "custom"
    custom_dir.mkdir()

    generate_addons(custom_dir, ['i', 'j', 'k'])

    env.context.custom_paths.add(custom_dir)

    assert len(env.modules.list(reload=True)) == 10
    assert len(env.addons_paths()) == 3

    for module in ['i', 'j', 'k', 'f', 'g', 'h']:
        assert env.modules.get(module).technical_name == module

    module_i = env.modules.get('i')
    assert module_i.path.exists() is True

    env.context.disabled_modules = {'i', 'j', 'k'}
    env.modules.remove_disabled()

    assert module_i.path.exists() is False

    for mod in env.modules.list(reload=True):
        mod.package()

   
def test_env_options(tmp_path):
    env = Environment()

    env.env_options()


def test_manifest_save_and_files(tmp_path):
    cur_dir = Path(tmp_path)

    env = Environment()
    env.context.custom_paths.add(cur_dir)

    generate_addons_full(
        cur_dir,
        'mod1',
        {'depends': ['web']},
        {
            "__init__.pyc": "some data",
            "models": {
                "__init__.pyc": "some data2",
            },
            "static": {
                "src": {
                    "index.html": "Index File"
                }
            }
        }
    )

    mod = env.modules.get('mod1')

    assert mod.technical_name == 'mod1'
    assert len([x for x in mod.files()]) == 2
    assert len([x for x in mod.static_assets()]) == 1


def test_manifest_from_path(tmp_path):
    cur_dir = Path(tmp_path)

    env = Environment()
    env.context.custom_paths.add(cur_dir)

    generate_addons_full(
        cur_dir,
        'mod1',
        {'depends': ['web']},
        {
            "static": {
                "src": {
                    "index.html": "Index File"
                }
            }
        }
    )

    mod = Manifest.from_path(cur_dir / 'mod1')

    assert mod.technical_name == 'mod1'
    assert mod.depends == ['web']


def test_manifest_requirements(tmp_path):
    cur_dir = Path(tmp_path)

    env = Environment()
    env.context.custom_paths.add(cur_dir)

    generate_addons_full(
        cur_dir,
        'mod1',
        {'depends': ['web']},
        {
            "static": {
                "src": {
                    "index.html": "Index File"
                }
            }
        }
    )

    mod = Manifest.from_path(cur_dir / 'mod1')
    assert mod.requirements() == set()

    mod.external_dependencies['python'] = ['cryptography']
    assert mod.requirements() == {'cryptography'}

    mod.save()
    mod = Manifest.from_path(cur_dir / 'mod1')
    assert mod.requirements() == {'cryptography'}
    assert mod.requirements({'cryptography': 'crypto'}) == {'crypto'}
    assert mod.requirements({'cryptography': ''}) == set()



from .models import ParentedObject


class BaseField(property):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        if 'default' in kwargs:
            self._default = kwargs['default']

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, owner, klass):
        # When trying to get the property from a class
        if not owner:
            return self

        if self.name not in owner._data and hasattr(self, '_default'):
            return self._default
        else:
            return owner._data.get(self.name)

    def __set__(self, owner, value):
        owner._data[self.name] = self.parse(owner, value)

    def parse(self, owner, data):
        return data


class String(BaseField):
    _default = ""

    def parse(self, owner, data):
        return str(data)


class Boolean(BaseField):
    _default = False

    # def __init__(self, *args, **kwargs):
    #    if 'default' not in kwargs:
    #        kwargs['default'] = False
    #    super().__init__(*args, **kwargs)

    def parse(self, owner, data):
        return bool(data)


class Object(BaseField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.object_class = kwargs['object_class']

    def parse_object(self, owner, data):
        if not data:
            data = {}

        obj = self.object_class.parse(data)

        if isinstance(obj, ParentedObject):
            obj.set_parent(owner)

        return obj

    def parse(self, owner, data):
        return self.parse_object(owner, data)


class ProxyObject(BaseField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getter = kwargs['getter']

    def parse(self, owner, data):
        return {
            "ref": data
        }

    def __get__(self, owner, klass):
        ref = super().__get__(owner, klass)
        if not ref:
            return None
        return getattr(owner, self.getter)(ref['ref'])


class List(Object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def parse(self, owner, data):
        if not data:
            return []

        vals = [
            (
                item
                if isinstance(item, self.object_class)
                else self.parse_object(owner, item)
            )
            for item in data
        ]

        for val in vals:
            if isinstance(val, ParentedObject):
                val.set_parent(owner)

        return vals


class Dict(List):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._key = kwargs['key']

    def parse(self, owner, data):
        if isinstance(data, dict):

            for val in data.values():
                if isinstance(val, ParentedObject):
                    val.set_parent(owner)

            return data

        objects = super().parse(owner, data)

        return {
            getattr(item, self._key): item
            for item in objects
        }


class Url(BaseField):
    pass



import os
import toml
import sys
import logging
from pathlib import Path
from contextlib import contextmanager
from configparser import NoSectionError, NoOptionError

from ..exceptions import OdooNotInstalled, NoConfigError
from ..compat import module_path
from ..modules.search import find_addons_paths, find_modules_paths
from ..utils import (
    filter_excluded_paths,
    to_path_list,
    convert_env_value,
    ConfigParser
)
from ..configuration.misc import (
    get_resource
)

from .context import Context
from .management import MangementApi
from .modules import ModuleApi


_logger = logging.getLogger(__name__)


class Environment(object):
    """
    Odoo Environment object.

    The odoo environment object is a container that store all
    the information required to prepare an environment for odoo.

    It can be used to browse the available modules in the configured
    odoo environment.

    Or to setup environment variables for the running odoo instance.

    It can also be used to configure the odoo.cfg file based on the
    current environment.

    Attributes:
        context (Context): The context to use

        env (Environment): The environment variables container


        requirement_file_path (Path): The path where to store the requirements
            file when merging pip requirements.

    """

    def __init__(
        self,
        context=None,
    ):
        """
        Initialize an environment.

        Parameters:
            context (Context): The context to use.
            env (Environment): The environment to use.
            strict_mode (bool): If the environment is strict.
        """
        if context is None:
            context = Context.from_env()

        self.context = context
        self.modules = ModuleApi(self)
        self.manage = MangementApi(self)
        self._config = None

    def path(self):
        """
        Returns the path where odoo is installed and can be imported.

        The base addons are installed in the folder `addons` relative
        to this path.
        """
        if self.context.odoo_base_path:
            return Path(self.context.odoo_base_path)

        try:
            path = module_path("odoo")
        except Exception:
            try:
                path = module_path("openerp")
            except Exception:
                raise OdooNotInstalled("Cannot find odoo base path")

        return path

    def check_odoo(self):
        try:
            self.path()
        except Exception:
            raise OdooNotInstalled(
                "Cannot use this api without odoo being installed"
            )

    def set_config(self, key, value, section='options'):
        """
        Set a configuration settings in the currently open configuration file.

        Example:

        .. code:: python

            env.set_config('server_wide_modules', 'web,base')
            env.set_config('max_handlers', '3', 'custom_section')

        Parameters:

            key (str): The name of the config to set.

            value (str): The value to store in the config

            section (str): The section to use in the ConfigParser object. By
                default, it uses the ``'options'`` section. This is the default
                section used by odoo. But the section can be set to something
                different to define additional sections for use.
        """
        if not self._config:
            raise NoConfigError("No config file currently open")

        self._config.set(section, key, value)

    def get_config(self, key, section='options'):
        """
        Get a configuration settings in the currently open configuration file.
        """
        if not self._config:
            raise NoConfigError("No config file currently open")

        try:
            return self._config.get(section, key)
        except NoOptionError:
            return None
        except NoSectionError:
            return None

    @contextmanager
    def config(self):
        """
        A context manager that can be used to read/write configuration for
        odoo.

        The context manager saves the configuration file when it is closed.


        .. code:: python

            with env.config():
                env.set_config('server_wide_modules', 'web,base')

        """
        config_path = Path(self.context.odoo_rc)

        if not self._config:
            nested = False
            conf = ConfigParser()
        else:
            nested = True
            conf = self._config

        if not self._config and config_path.exists():
            try:
                conf.read(str(config_path))
            except Exception:
                _logger.info("Couldn't read ODOO_RC file.", exc_info=True)

        try:
            self._config = conf
            yield conf
        except Exception:
            if not nested:
                self._config = None
            raise
        finally:
            if not nested:
                self._config = None

        if not nested:
            try:
                with config_path.open('w') as out:
                    conf.write(out)
            except Exception:
                _logger.info("Couldn't write config ", exc_info=True)
                with config_path.open('wb') as out:
                    conf.write(out)

    def addons_paths(self):
        """
        Returns the addons path configured for this environment.

        For example, it would find the path addons in the odoo installed
        folder. And in custom paths defined to search for addons.

        If an addons path had modules in ``/a/b``, ``/a/d/e/f`` and ``/a/b/c``.
        It would return the following list of addons paths.

        .. code:: python

            ['/a/b', '/a/b/c', '/a/d/e/f']

        Odoo doesn't load folders recursively so if you have modules within
        folder that also contains modules. The addons_paths have to be defined.

        With this, you can store your modules in to ``'/addons'`` and only
        define as custom_paths ``'/addons'``. If there are modules in
        ``'/addons/**'``.  Those will be returned by ``addons_paths()``.


        .. code:: python

            addons_paths = ",".join(env.addons_paths())

            with env.config():
                env.set_config("addons_paths", addons_paths)

        Returns:
            paths (List<Path>): The list of paths containing installable
                addons.
        """
        try:
            with self.config() as config:
                paths = config.get('options', 'addons_paths')

            config_paths = set(
                Path(path)
                for path in paths.split(',')
                if path
            )

            if len(config_paths) > 0 and not self.context.force_addons_lookup:
                return config_paths

        except Exception:
            config_paths = set()

        base_addons_paths = config_paths

        try:
            if not self.context.exclude_odoo:
                base_addons = self.path() / "addons"
                base_addons_paths.add(base_addons)
        except OdooNotInstalled:
            pass

        base_addons_paths |= self.context.custom_paths

        orig_valid_paths = find_addons_paths(
            base_addons_paths,
            options=self.context
        )

        if self.context.excluded_paths:
            excluded_paths = to_path_list(self.context.excluded_paths)
            valid_paths = filter_excluded_paths(
                orig_valid_paths, excluded_paths
            )
        else:
            valid_paths = orig_valid_paths

        return valid_paths

    def odoo_config(self):
        try:
            from odoo.tools import config as odoo_config
            return odoo_config
        except ImportError:
            try:
                from openerp.tools import config as odoo_config
                return odoo_config
            except (ImportError, ModuleNotFoundError):
                raise OdooNotInstalled(
                    "Cannot use config without odoo being installed"
                )

    def env_options(self):
        try:
            odoo_config = self.odoo_config()
        except OdooNotInstalled:
            return {}

        params_by_name = {}
        for key, opt in odoo_config.casts.items():
            value_opt = opt.get_opt_string()
            value_opt = value_opt.upper().replace('--', 'ODOO_')
            value_opt = value_opt.replace('-', '_')
            params_by_name[value_opt] = key

        configs = {}

        for key, value in os.environ.items():
            if not key.startswith('ODOO_'):
                continue

            if key in params_by_name:
                config_name = params_by_name[key]
                converted_value = convert_env_value(key, value)
                configs[config_name] = converted_value

        return configs

    def requirement_files(self, lookup_requirements=False):
        found_files = set()

        for cur_path in self.addons_paths():
            if lookup_requirements:
                found_files |= set(cur_path.glob("**/requirements.txt"))
            else:
                requirement_file = cur_path / 'requirements.txt'
                if requirement_file.exists():
                    found_files.add(requirement_file)

        return found_files

    def _default_package_map(self):
        version = self.odoo_version()

        if not version:
            return {}

        file_path = "packages/map-{version}.toml".format(
            version=version
        )

        resource_path = get_resource('odoo_tools', file_path)

        if not resource_path.exists():
            return {}

        with resource_path.open('r') as fin:
            data = toml.load(fin)

        return data

    def package_map(self):
        """
        Returns a package map.

        The package map is used to map some module name
        to python package names.

        By default, in newer versions of odoo, it will check
        for the package name. But unfortunately some modules
        will still have the module name defined in their
        python external dependencies.

        When requirements are built from the odoo modules available
        in the odoo environment. It will wrongly attempt to install
        let say the module "ldap". ldap is the name of the module
        that can be imported, but the package that needs to be
        installed is python-ldap.

        Such map could look like this:

        .. code-block:: python

            {'ldap': 'python-ldap'}

        Then when the packages required are found, they can be mapped
        to those package names to find the exact python package name.

        Mapping a name to an empty string would remove the name from
        the requirements. This can be useful to remove package defined
        in the external dependencies that aren't actual packages but
        builtin dependencies that would be already part of python
        itself.

        .. code-block:: python

            {'asyncio': ''}

        The behaviour could be also useful when you want to install an
        alternative to let say the barcode module.

        Returns:
            dict: Key, Value of mapped module/package name.
        """
        base_map = self._default_package_map()

        package_map_file = self.context.package_map_file
        if not package_map_file:
            return base_map

        package_map_path = Path(package_map_file)

        if not package_map_path.exists():
            return base_map

        content = package_map_path.open('r').read()
        package_map = toml.loads(content)

        new_vals = {
            key.lower(): value.lower()
            for key, value in package_map.items()
        }

        base_map.update(new_vals)
        return base_map

    def odoo_version(self):
        """
        Returns the odoo version.

        In case odoo isn't installed, it will fallback to the
        context variable `odoo_version` which can be set through
        environment variable ODOO_VERSION.

        Returns:
            int|None: The major version number or None
        """
        try:
            from odoo.release import version_info
            return version_info[0]
        except ImportError:
            try:
                return int(float(self.context.odoo_version))
            except Exception:
                return None



import six
import logging

from .db import DbApi
from ..entrypoints import execute_entrypoint
from ..configuration.odoo import (
    OfficialRelease,
    GitRelease
)


_logger = logging.getLogger(__name__)


class MangementApi(object):
    def __init__(self, environment):
        self.environment = environment
        self._initialized = False

    def db(self, database):
        """
        Returns a DbApi instance.

        Args:
            database (str): The database name

        Returns:
            DbApi: The db api object to manage odoo with the
                provided database name.
        """
        return DbApi(
            self,
            database
        )

    @property
    def config(self):
        """
        Returns and odoo configparser.

        This method is a simple shortcut around the odoo config.
        In case odoo isn't installed, it will simply raise an
        exception that can be handled instead of having to
        handle in your code to handle if odoo can be imported.

        Raises:
            SystemError: If odoo cannot be imported.

        Returns:
            configparser: the original odoo config object.
        """
        try:
            from odoo.tools import config
            return config
        except Exception:
            pass

        raise SystemError("Odoo doesn't seems to be installed")

    def initialize_odoo(self):
        """
        Initialize the odoo environment.

        This method will load entrypoints that can be declared in python
        packages or by explicitly creating them using the entrypoints
        utilities in odoo-tools.

        The first entrypoint to be called is:
        `odoo_tools.manage.before_config`. This entrypoint is supposed
        to be used to prepare the odoo configuration by filling things
        like addons paths etc. Then the entrypoint
        `odoo_tools.manage.initialize_odoo` is called. This one is responsible
        for loading the rest of odoo, for example it would setup the AddonsHook
        to load `odoo.addons` modules. Or to initialize the odoo logger. Then
        the `odoo_tools.manage.after_config` would be called to load antyhing
        that you'd want after the odoo internal configuration is effectly
        loaded.

        By default the entrypoints are undefined, this way you can
        adjust the way you want to have odoo loaded before starting
        http workers for example.
        """
        self._initialized = True
        execute_entrypoint("odoo_tools.manage.before_config", self)
        execute_entrypoint("odoo_tools.manage.initialize_odoo", self)
        execute_entrypoint("odoo_tools.manage.after_config", self)

    def install_odoo(
        self,
        version,
        release=None,
        ref=None,
        repo='https://github.com/odoo/odoo.git',
        opts=None
    ):
        """
        Install odoo in the current python environment.

        Unfortunately, odoo doesn't provide a standard way to setup odoo
        through pip using pypi repositories. As a result, many companies end
        up installing odoo in many different ways. Some ways are better than
        others, but most of those methods can't be considered best practices.

        Implementing odoo's installation within this library is better in some
        ways as other methods available.

        In short, it can currently install odoo from 2 types of sources.

        Release Mode

        The release it the release
        name used in `Nightly builds <NIGHTLY>`_ of Odoo. It uses the format
        of the date of the build that you need.

        Git Mode

        The git mode will take a reference and a repository that needs to be
        fetched. The repository defaults to the odoo repository, but an
        alternative repository can be used if necessary.

        Regardless of the mode, each mode can have their file cached in an
        alternate folder. So you can try to cache the fetch in a CI environment
        to speed up odoo installation. When odoo is installed, it will pre
        process the odoo environment to provide a consistant setup. By default,
        odoo will only install addons located in odoo/addons. But in the git
        repository, addons are located in /addons and /odoo/addons. If you want
        to install odoo with pip manually, you'd have to move all addons in the
        root to the odoo/addons folder. This library does that for you.

        It's also possible to strip language translations out of the source
        folder for odoo. Most people really don't need to keep all translation
        files in their environment. It's not a big issue, but it can make
        docker images drastically smaller when translations aren't installed
        with odoo.

        This can save a non negligable amount of storage and time to upload
        large images.

        Then the source code used to install odoo gets automatically cleaned
        up to save even more space as .git data aren't necessary in a
        statically built environment. Ironically, installing odoo from scratch
        without cache can be a lot faster than loading the whole repository in
        a cache then doing incremental fetch to keep the source up to date.

        .. _NIGHTLY: https://nightly.odoo.com/
        """
        if hasattr(opts, 'cache'):
            cache = opts.cache
        else:
            cache = None

        if release:
            installer = OfficialRelease(
                version,
                release,
                options=opts,
                cache=cache
            )
        else:
            installer = GitRelease(
                version,
                repo,
                ref or version,
                options=opts,
                cache=cache
            )

        with installer:
            if installer.need_update():
                installer.fetch()
                installer.checkout()
                installer.install()

    def packages(self):
        """
        It will lookup all the native packages that needs to be installed.

        Any file matching the name apt-packages.txt will be merge together
        to form the list of packages that needs to be installed.

        For example, if you have a module that requires git to be installed.
        You'd create a file with the following content:

        .. code-block::

            git

        Then when calling this method with properly configured addons_paths,
        it will automatically find this file merge it with other files
        and then you'll be able to use the output of this method and call
        the package manager of your choice with this list of package to
        install.

        Returns:
            list(str): list of package name.
        """
        package_list = set()
        paths = self.environment.addons_paths()

        _logger.info(
            "Looking up for packages in %s",
            paths
        )

        for addons_path in paths:
            for packages in addons_path.glob('**/apt-packages.txt'):
                _logger.info("Installing packages from %s", packages)

                with packages.open('r') as pack_file:
                    lines = [
                        six.ensure_text(line).strip()
                        for line in pack_file
                        if six.ensure_text(line).strip()
                    ]

                    package_list.update(set(lines))

        context = self.environment.context

        _logger.info(
            "Adding extra packages %s",
            context.extra_apt_packages
        )

        for package in context.extra_apt_packages:
            package_list.add(
                six.ensure_text(package).strip()
            )

        return package_list



import pytest
from odoo_tools.api.objects import CompanySpec, Manifest


def test_company_spec():
    company = CompanySpec(country_code='CA')
    assert company.country_code == 'CA'


def test_custom_empty_manifest(tmp_path):
    manifest = Manifest(tmp_path / 'sup')

    assert manifest.technical_name == 'sup'

    assert manifest.depends == []
    assert manifest.demo == []
    assert manifest.data == []
    assert manifest.application is False
    assert manifest.installable is True
    assert manifest.version == "0.0.0"
    assert manifest.external_dependencies == dict()

    assert manifest._manifest_file == manifest.path / '__manifest__.py'
    assert manifest.static_assets() == []

    manifest.name = "some name"

    manifest.save()

    manifest = Manifest.from_path(manifest.path)
    assert manifest.name == "some name"


def test_openerp_manifest(tmp_path):
    description = """
Some module
===========

some module description
"""

    manifest = {
        "name": "module_name",
        "summary": "summary",
        "depends": ['web', 'base'],
        "description": description
    }

    module_path = tmp_path / 'mod'
    module_path.mkdir()

    html_file = module_path / 'static/description/index.html'

    html_file.parent.mkdir(parents=True)

    with html_file.open('w') as index:
        index.write("""<!doctype html>
<html>
<head>
    <link src="style.css" rel="stylesheet" type="text/css" />
</head>
<body>
    <div></div>
</body>
</html>
""")

    manifest_file = module_path / '__openerp__.py'

    with manifest_file.open('w') as fout:
        fout.write(repr(manifest))

    manifest = Manifest.from_path(module_path, render_description=True)

    assert manifest.name == "module_name"
    assert manifest.summary == "summary"
    assert manifest.depends == ['web', 'base']
    assert manifest.installable is True

    manifest.disable()

    assert manifest.installable is False

    manifest.save()

    manifest = Manifest.from_path(module_path, render_description=True)
    assert manifest.installable is False


def test_manifest_set_attribute(tmp_path):
    manifest = Manifest(tmp_path / 'sup')

    manifest.set_attribute(
        ['external_dependencies', 'python'],
        ['web', 'base', 'fun']
    )

    deps = manifest.external_dependencies['python']
    assert deps == ['web', 'base', 'fun']


def test_invalid_manifest(tmp_path):
    manifest = Manifest(tmp_path / 'sup')
    manifest.save()

    with manifest._manifest_file.open('wb') as fout:
        fout.write('{"depends": base, web}'.encode())

    with pytest.raises(SyntaxError):
        manifest = Manifest.from_path(manifest.path)
